# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import namedtuple

from data_pipeline.schematizer_clientlib.models.model_base import BaseModel


"""
Represent the data of a namespace.  Namespace is a group which the avro schemas
belong to.  It is the highest grouping level of schemas.  For example,
`yelp_main` could be a namespace.

Args:
    namespace_id (int): The id of the namespace.
    name (str): The name of the namespace.
"""
Namespace = namedtuple('Namespace', ['namespace_id', 'name'])


class _Namespace(BaseModel):

    def __init__(self, namespace_id, name):
        self.namespace_id = namespace_id
        self.name = name

    @classmethod
    def from_response(cls, response):
        return cls(
            namespace_id=response.namespace_id,
            name=response.name
        )

    def to_result(self):
        return Namespace(
            namespace_id=self.namespace_id,
            name=self.name
        )



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from data_pipeline.tools.introspector.models.base import BaseIntrospectorModel


class IntrospectorSource(BaseIntrospectorModel):
    def __init__(self, source_obj, active_sources=None):
        super(IntrospectorSource, self).__init__(
            source_obj
        )
        self._fields = [
            'name', 'source_id', 'owner_email', 'namespace'
        ]
        self.namespace = source_obj.namespace.name
        # Need to check for none in case of empty list
        if active_sources is not None:
            self._fields.append('active_topic_count')
            active_source = active_sources.get(source_obj.source_id, None)
            self.active_topic_count = 0 if (
                not active_source
            ) else active_source['active_topic_count']



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import math
import os
import socket
import time

import simplejson
from cached_property import cached_property

from data_pipeline._kafka_producer import LoggingKafkaProducer
from data_pipeline.config import get_config
from data_pipeline.expected_frequency import ExpectedFrequency
from data_pipeline.message import MonitorMessage
from data_pipeline.registrar import Registrar
from data_pipeline.schematizer_clientlib.schematizer import get_schematizer
from data_pipeline.team import Team


logger = get_config().logger


class Client(object):
    """The client superclasses both the Producer and Consumer, and is largely
    responsible for shared producer/consumer registration responsibilities and
    monitoring.

    Note:

        Client will be responsible for producer/consumer registration,
        which will be implemented in DATAPIPE-1154.

    Args:
        client_name (str): Name associated with the client - this name will
            be used as the client's identifier for producer/consumer
            registration.  This identifier should be treated as a constant
            once defined, and should be unique per use case and application.

            For example, if a team has a service with two consumers, each
            filling a different role, and a consumer in yelp main, all three
            should have different client names so their activity can be
            differentiated.  Multiple instances of a producer or consumer in the
            same application should share a client name.  If an application is
            both consuming and producing for the same purpose (i.e. transforming
            and republishing a topic), the consumer and producer should share a
            client name, otherwise the client name should be unique across
            logical applications.

            There is no need to use different client names in different
            environments, but we do suggest namespacing.  For example,
            `services.yelp-main.datawarehouse.rich-transformers` or
            `services.user_tracking.review_tracker`.
        team_name (str): Team name, as defined in `sensu_handlers::teams` (see
            y/sensu-teams).  notification_email must be defined for a team to be
            registered as a producer or consumer.  This information will be used
            to communicate about data changes impacting the client.

            `sensu_handlers::teams` is the canonical data source for team
            notifications, and its usage was recommended by ops.  It was also
            adopted for usage in ec2 instance tagging in y/cep427.
        expected_frequency_seconds (int, ExpectedFrequency): How frequently, in seconds,
            that the client expects to run to produce or consume messages.
            Any positive integer value can be used, but some common constants
            have been defined in
            :class:`data_pipeline.expected_frequency.ExpectedFrequency`.

            See :class:`data_pipeline.expected_frequency.ExpectedFrequency` for
            additional detail.
        monitoring_enabled (Optional[bool]): If true, monitoring will be enabled
            to record client's activities. Default is true.
        dry_run (Optional[bool]): If true, client will skip publishing message to
            kafka. Default is false.

    Raises:
        InvalidTeamError: If the team specified is either not defined or does
            not have a notification_email registered.  Ops deputies can modify
            `sensu_handlers::teams` in puppet to add a team.
    """

    def __init__(
        self,
        client_name,
        team_name,
        expected_frequency_seconds,
        monitoring_enabled=True,
        dry_run=False
    ):
        self.monitor = _Monitor(
            client_name,
            self.client_type,
            start_time=_Monitor.get_monitor_window_start_timestamp(
                time.time()
            ),
            monitoring_enabled=monitoring_enabled,
            dry_run=dry_run
        )
        self.client_name = client_name
        self.team_name = team_name
        self.expected_frequency_seconds = expected_frequency_seconds
        self.registrar = Registrar(
            team_name,
            client_name,
            self.client_type,
            self.expected_frequency_seconds
        )

    @property
    def client_name(self):
        """Name associated with the client"""
        return self._client_name

    @client_name.setter
    def client_name(self, client_name):
        if not client_name or not isinstance(client_name, (str, unicode)):
            raise ValueError("Client name must be non-empty text")
        self._client_name = client_name

    @property
    def team_name(self):
        """Team associated with the client"""
        return self._team_name

    @team_name.setter
    def team_name(self, team_name):
        if not Team.exists(team_name):
            raise ValueError(
                "Team name must exist: see the team_name argument at "
                "y/dp_client_docs for detailed information about adding a team."
            )
        self._team_name = team_name

    @property
    def expected_frequency_seconds(self):
        """How frequently, in seconds, that the client expects to run to produce
        or consume messages.
        """
        return self._expected_frequency_seconds

    @expected_frequency_seconds.setter
    def expected_frequency_seconds(self, expected_frequency_seconds):
        if isinstance(expected_frequency_seconds, ExpectedFrequency):
            expected_frequency_seconds = expected_frequency_seconds.value

        if not (isinstance(expected_frequency_seconds, int) and expected_frequency_seconds >= 0):
            raise ValueError("Client name must be non-empty text")
        self._expected_frequency_seconds = expected_frequency_seconds

    @property
    def client_type(self):
        """String identifying the client type."""
        raise NotImplementedError


class _Monitor(object):
    """The class implements functionality of monitoring messages which are
    published/consumed by clients using kafka.

    In current implementation, monitoring_window_in_sec is stored in the staticconf
    and is the same for every client. As one of the future features, one could add
    the option to have a custom monitoring_window_in_sec for every client and choose
    the appropriate duration based on the average number of messages processed by
    the client.

    Args:
        client_name (str): name of the associated client.
        client_type (str): type of the client the _Monitor is associated to.
            Could be either producer or consumer.
        start_time (int): unix start_time when _Monitor will start tracking
            the number of messages produced/consumed by the associated client.
        monitoring_enabled (Optional[bool]): If true, monitoring will be enabled
            to record client's activities. Default is true.
        dry_run (Optional[bool]): If true, client will skip publishing message to
            kafka. Default is false.
    """

    def __init__(
        self,
        client_name,
        client_type,
        start_time=0,
        monitoring_enabled=True,
        dry_run=False
    ):
        self.client_name = client_name
        self.client_type = client_type

        self.monitoring_enabled = monitoring_enabled
        if not self.monitoring_enabled:
            return

        self.topic_to_tracking_info_map = {}
        self._monitoring_window_in_sec = get_config().monitoring_window_in_sec
        self.start_time = start_time
        self.producer = LoggingKafkaProducer(
            self._notify_messages_published,
            dry_run=dry_run
        )
        self.dry_run = dry_run
        self._last_msg_timestamp = None

    @classmethod
    def get_monitor_window_start_timestamp(cls, timestamp):
        _monitor_window_in_sec = get_config().monitoring_window_in_sec
        return (int(math.floor(int(timestamp) / _monitor_window_in_sec)) *
                _monitor_window_in_sec)

    def _get_default_record(self, topic):
        """Returns the default version of the topic_to_tracking_info_map entry
        """
        return {
            'topic': topic,
            'client_type': self.client_type,
            'message_count': 0,
            'start_timestamp': self.start_time,
            'host_info': socket.gethostname(),
            'client_name': self.client_name
        }

    def _get_record(self, topic):
        """returns the record associated with the topic for any client
        If the topic has no record, a new record is created and returned
        """
        tracking_info = self.topic_to_tracking_info_map.get(topic)
        if tracking_info is None:
            tracking_info = self._get_default_record(topic)
            self.topic_to_tracking_info_map[topic] = tracking_info
        return tracking_info

    def _publish(self, tracking_info):
        """publish monitoring results, stored in the monitoring_message, using
        the producer
        """
        self.producer.publish(
            MonitorMessage(
                schema_id=self.monitor_schema_id,
                payload_data=tracking_info,
                dry_run=self.dry_run
            )
        )

    @cached_property
    def monitor_schema(self):
        return get_schematizer().register_schema(
            namespace=self._monitor_schema['namespace'],
            source=self._monitor_schema['name'],
            schema_str=simplejson.dumps(self._monitor_schema),
            source_owner_email='bam+data_pipeline+monitor@yelp.com',
            contains_pii=False
        )

    @cached_property
    def _monitor_schema(self):
        schema_file = os.path.join(
            os.path.dirname(__file__),
            'schemas/monitoring_message_v1.avsc'
        )
        with open(schema_file, 'r') as f:
            schema_string = f.read()
        return simplejson.loads(schema_string)

    @cached_property
    def monitor_schema_id(self):
        return self.monitor_schema.schema_id

    @cached_property
    def monitor_topic(self):
        return str(self.monitor_schema.topic.name)

    def _reset_monitoring_record(self, tracking_info):
        """resets the record for a particular topic by resetting
        message_count to 0 and updating the start_timestamp to the
        starting point of the next monitoring_window
        """
        tracking_info['message_count'] = 0
        tracking_info['start_timestamp'] += self._monitoring_window_in_sec

    def _notify_messages_published(self, position_data):
        """Called to notify the client of successfully published messages.

        Args:
            position_data (:class:PositionData): PositionData structure
                containing details about the last messages published to Kafka,
                including Kafka offsets and upstream position information.
        """
        logger.info("Client: " + self.client_name + " published monitoring message")

    def record_message(self, message):
        """Used to handle the logic of recording monitoring_message in kafka
        and resetting it if necessary.
        """
        if not self.monitoring_enabled:
            return

        self._last_msg_timestamp = message.timestamp

        track_info = self._get_record(message.topic)
        track_info = self._flush_previous_track_info(track_info)
        track_info['message_count'] += 1

    def _flush_previous_track_info(self, current_track_info):
        next_start_time = (current_track_info['start_timestamp'] +
                           self._monitoring_window_in_sec)
        while next_start_time <= self._last_msg_timestamp:
            self._publish(current_track_info)
            self._reset_monitoring_record(current_track_info)
            next_start_time = (current_track_info['start_timestamp'] +
                               self._monitoring_window_in_sec)
        return current_track_info

    def flush_buffered_info(self):
        """Publishes the buffered information, stored in topic_to_tracking_info_map,
        to kafka and resets topic_to_tracking_info_map to an empty dictionary
        """
        if not self.monitoring_enabled:
            return

        for track_info in self.topic_to_tracking_info_map.values():
            last_track_info = self._flush_previous_track_info(track_info)
            self._publish(last_track_info)
        self.producer.flush_buffered_messages()
        self.topic_to_tracking_info_map = {}

    def close(self):
        """Called when the associated client is exiting/closing. Calls flush_buffered_info
        and also closes monitoring.producer.
        """
        if not self.monitoring_enabled:
            return

        self.flush_buffered_info()
        self.producer.close()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import glob


def get_file_paths_from_glob_patterns(glob_patterns):
    """ Return a set of files matching the given list of glob patterns
     (for example ["./test.sql", "./other_tables/*.sql"])
    """
    file_paths = set()
    for glob_pattern in glob_patterns:
        file_paths |= set(glob.glob(glob_pattern))
    return file_paths



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import namedtuple

from data_pipeline.schematizer_clientlib.models.model_base import BaseModel


"""
Represent the data of meta attribute mapping store. Meta attribute namespace
mapping should represent a mapping of a namespace and the corresponding meta
attribute schema id.
"""
MetaAttributeNamespaceMapping = namedtuple(
    'MetaAttributeNamespaceMapping',
    ['namespace_id', 'meta_attribute_schema_id']
)


class _MetaAttributeNamespaceMapping(BaseModel):
    """Internal class used to convert from/to various data structure and
    facilitate constructing the return value of schematizer functions.
    """

    def __init__(self, namespace_id, meta_attribute_schema_id):
        self.namespace_id = namespace_id
        self.meta_attribute_schema_id = meta_attribute_schema_id

    @classmethod
    def from_response(cls, namespace_id, meta_attribute_schema_id):
        return cls(
            namespace_id=namespace_id,
            meta_attribute_schema_id=meta_attribute_schema_id
        )

    def to_result(self):
        return MetaAttributeNamespaceMapping(
            namespace_id=self.namespace_id,
            meta_attribute_schema_id=self.meta_attribute_schema_id
        )



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import simplejson

from data_pipeline.tools.introspector.list_command.base_command import _BaseListCommand


class TopicsListCommand(_BaseListCommand):

    list_type = 'topics'
    fields = [
        'name', 'topic_id', 'contains_pii',
        'primary_keys',
        'in_kafka', 'message_count',
        'source_name', 'source_id',
        'namespace',
        'created_at', 'updated_at'
    ]

    @classmethod
    def add_parser(cls, subparsers):
        list_command_parser = subparsers.add_parser(
            "topics",
            description=cls.get_description(),
            add_help=False
        )

        cls.add_base_arguments(list_command_parser)
        cls.add_source_and_namespace_arguments(list_command_parser)

        list_command_parser.set_defaults(
            command=lambda args:
                cls("data_pipeline_introspector_list").run(args, list_command_parser)
        )

    def process_args(self, args, parser):
        super(TopicsListCommand, self).process_args(args, parser)
        self.process_source_and_namespace_args(args, parser)

    def run(self, args, parser):
        self.process_args(args, parser)
        print simplejson.dumps(self.list_topics(
            source_id=self.source_id,
            namespace_name=self.namespace,
            source_name=self.source_name,
            sort_by=self.sort_by,
            descending_order=self.descending_order
        ))



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import time
from datetime import datetime

from cached_property import cached_property

from data_pipeline.schematizer_clientlib.schematizer import get_schematizer


class ConsumerSource(object):
    """Base class to specify the Kafka topics where the consumer would like to
    fetch messages from.
    """

    def get_topics(self):
        """Get a list of topics where the consumer get the messages from.  The
        derived class must implement this function.  It should return a list
        of topic names.

        Returns:
            List[str]: A list of topic names.  If no topics, an empty list is
            returned.
        """
        raise NotImplemented()

    @cached_property
    def schematizer(self):
        return get_schematizer()


class FixedTopics(ConsumerSource):
    """Consumer tails one or a fixed set of topics.

    Args:
        topic_names: Variable number of Kafka topic names to consume from.
            At least one topic must be specified.
    """

    def __init__(self, *topic_names):
        if not any(topic_names):
            raise ValueError("At least one topic must be specified.")
        self.topics = topic_names

    def get_topics(self):
        return self.topics


class TopicsInFixedNamespaces(ConsumerSource):
    """Consumer tails all the topics in specified namespaces.

    Args:
        namespace_names (tuple(str)): Variable number of namespace names in which all the
        topics will be tailed by the consumer.
    """

    def __init__(self, *namespace_names):
        if not any(namespace_names):
            raise ValueError("At least one namespace must be specified.")
        self.namespace_names = namespace_names

    def get_topics(self):
        topic_names = []
        for namespace_name in self.namespace_names:
            topic_names.extend(
                topic.name
                for topic in self.schematizer.get_topics_by_criteria(
                    namespace_name=namespace_name
                )
            )
        return topic_names


class TopicInSource(ConsumerSource):
    """Consumer tails the topics of specified source.  The source is specified
    by giving the namespace name and the source name stored in the Schematizer.
    For example, to tail all the topics associated to yelp-main biz table, the
    namespace name could be something like 'yelp` and the source name could be
    'biz'.

    Args:
        namespace_name (str): Namespace name of the specified source in which
            all the topics will be tailed by the consumer.
        source_name (str): Source name of the specified source in which all the
            topics will be tailed by the consumer.
    """

    def __init__(self, namespace_name, source_name):
        if not namespace_name:
            raise ValueError("namesapce_name must be specified.")
        if not source_name:
            raise ValueError("source_name must be specified.")
        self.namespace_name = namespace_name
        self.source_name = source_name

    def get_topics(self):
        topics = self.schematizer.get_topics_by_criteria(
            namespace_name=self.namespace_name,
            source_name=self.source_name
        )
        return [topic.name for topic in topics]


class FixedSchemas(ConsumerSource):
    """Consumer tails the topics given a list of schema ids. FixedSchema
    consumer source has `schema_to_topic_map` map that provides mapping
    from schema_id to its corresponding topic name.

    Args:
        schema_ids (tuple(int)): Variable number of schema IDs of avro schemas
        registered in the Schematizer.
    """

    def __init__(self, *schema_ids):
        if not any(schema_ids):
            raise ValueError("At least one schema id must be specified.")
        self.schema_ids = schema_ids

    def get_topics(self):
        topics = {
            self.schematizer.get_schema_by_id(schema_id).topic.name
            for schema_id in self.schema_ids
        }
        return list(topics)

    def get_schema_to_topic_map(self):
        schema_to_topic_map = {
            schema_id: self.schematizer.get_schema_by_id(schema_id).topic.name
            for schema_id in self.schema_ids
        }
        return schema_to_topic_map


class TopicInDataTarget(ConsumerSource):
    """Consumer tails the topics in the specified data target.

    Args:
        data_target_id (int): the ID of the data target stored in the Schematizer.
    """

    def __init__(self, data_target_id):
        if not data_target_id:
            raise ValueError("data_target_id must be specified.")
        self.data_target_id = data_target_id

    def get_topics(self):
        topics = self.schematizer.get_topics_by_data_target_id(self.data_target_id)
        return [topic.name for topic in topics]


class NewTopicsOnlyInFixedNamespaces(TopicsInFixedNamespaces):
    """Consumer tails all the topics in specified namespaces, but
    it internally keeps track of the previous query timestamp for each
    namespace and only returns the topics created after the last query
    timestamp, including the topics created at the last query timestamp for
    the namespace. It means the same topics returned previously may be included
    again if their created_at timestamp is right at previous query timestamp.

    Args:
        namespace_names (tuple(str)): Variable number of namespace names in which all the
        topics will be tailed by the consumer.
    """

    def __init__(self, *namespace_names):
        super(NewTopicsOnlyInFixedNamespaces, self).__init__(*namespace_names)
        self.last_query_timestamp = {}

    def get_topics(self):
        topic_names = []
        for namespace_name in self.namespace_names:
            created_after_timestamp = long(time.time())
            topic_names.extend(
                topic.name
                for topic in self.schematizer.get_topics_by_criteria(
                    namespace_name=namespace_name,
                    created_after=self.last_query_timestamp.get(namespace_name)
                )
            )
            self.last_query_timestamp[namespace_name] = created_after_timestamp
        return topic_names


class NewTopicOnlyInSource(TopicInSource):
    """Consumer tails the topics of the specified source, but it internally
    keeps track of the previous query timestamp and only returns the topics
    created after the last query timestamp, including the topics created at the
    last query timestamp.  It means the same topics returned previously may be
    included again if their created_at timestamp is right at previous query
    timestamp.

    Args:
        namespace_name (str): Namespace name of the specified source in which
            all the topics will be tailed by the consumer.
        source_name (str): Source name of the specified source in which all the
            topics will be tailed by the consumer.
    """

    def __init__(self, namespace_name, source_name):
        super(NewTopicOnlyInSource, self).__init__(
            namespace_name,
            source_name
        )
        self.last_query_timestamp = None

    def get_topics(self):
        topics = self.schematizer.get_topics_by_criteria(
            namespace_name=self.namespace_name,
            source_name=self.source_name,
            created_after=self.last_query_timestamp
        )
        self.last_query_timestamp = long(time.time())
        return [topic.name for topic in topics]


class NewTopicOnlyInDataTarget(TopicInDataTarget):
    """Consumer tails the topics in the specified data target, but it internally
    keeps track the previous query timestamp and only returns the topics created
    after the last query timestamp, including the topics created at the last
    query timestamp.  It means the same topics returned previously may be included
    again if their created_at timestamp is right at previous query timestamp.

    Args:
        data_target_id (int): the ID of the data target stored in the Schematizer.
    """

    def __init__(self, data_target_id):
        super(NewTopicOnlyInDataTarget, self).__init__(data_target_id)
        self.last_query_timestamp = None

    def _is_topic_created_after_last_query(self, topic):
        return (
            topic.created_at >= datetime.fromtimestamp(
                self.last_query_timestamp,
                topic.created_at.tzinfo
            )
        )

    def get_topics(self):
        topics = self.schematizer.get_topics_by_data_target_id(
            self.data_target_id
        )
        topic_names = [
            topic.name for topic in topics
            if (not self.last_query_timestamp or
                self._is_topic_created_after_last_query(topic)
                )
        ]
        self.last_query_timestamp = long(time.time())
        return topic_names



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import logging
import os

import staticconf
from bravado.client import SwaggerClient
from cached_property import cached_property
from kafka_utils.util.config import ClusterConfig


namespace = 'data_pipeline'
data_pipeline_conf = staticconf.NamespaceReaders(namespace)


class Config(object):
    """Contains configuration data for the clientlib.

    Configuration can be adjusted using staticconf.

    Example::

        When using the clientlib in a service, the minimum recommended
        configuration is::

            module_config:
                ...
                - namespace: smartstack_services
                  file: /nail/etc/services/services.yaml
                - namespace: data_pipeline
                  config:
                    load_schematizer_host_and_port_from_smartstack: True
                    kafka_cluster_type: datapipe
                    kafka_cluster_name: uswest2-devc
                    skip_messages_with_pii: True
                - namespace: clog
                  file: /nail/srv/configs/clog.yaml
    """
    YOCALHOST = '169.254.255.254'

    @cached_property
    def logger(self):
        """Logger instance for the clientlib"""
        return logging.getLogger('data_pipeline_clientlib')

    @property
    def should_use_testing_containers(self):
        """Used as a config that will not be overwritten in tests where
        something like load_schematizer_host_and_port_from_smartstack may be.

        Overwrite this configuration for testing purposes only.

        You may also set this configuration manually when starting a test run
        similarly to how it's done in testing_helpers.containers
        """
        return data_pipeline_conf.read_string(
            'should_use_testing_containers',
            default=None
        )

    @property
    def schematizer_port(self):
        """Port for the schematizer as an int.

        If :meth:`load_schematizer_host_and_port_from_smartstack` is True, this
        value will be loaded from smartstack instead of read directly.
        """
        return int(self.schematizer_host_and_port.split(':')[-1])

    @property
    def schematizer_host_and_port(self):
        """Host and port for the schematizer, in the format `host:port`.


        If :meth:`load_schematizer_host_and_port_from_smartstack` is True, this
        value will be loaded from smartstack instead of read directly.
        """
        if (self.load_schematizer_host_and_port_from_smartstack and
                not self.should_use_testing_containers):
            from yelp_servlib.config_util import get_service_host_and_port  # NOQA
            host, port = get_service_host_and_port('schematizer.main')
            return "{0}:{1}".format(host, port)
        else:
            return data_pipeline_conf.read_string(
                'schematizer_host_and_port',
                default='{0}:49256'.format(self.YOCALHOST)
            )

    @property
    def topic_refresh_frequency_seconds(self):
        """The frequency how often the Consumer refreshes the consumer source
        topics from the Schematizer.  The frequency is specified in seconds,
        default to 300 seconds (5 minutes).  The Consumer will automatically
        start consuming messages from newly picked-up topics.
        """
        return data_pipeline_conf.read_float('topic_refresh_frequency_seconds', 300)

    @property
    def load_schematizer_host_and_port_from_smartstack(self):
        """Load the host and port from SmartStack instead of setting it
        directly.

        SmartStack must be configured with staticconf in order to use this
        option.  For information on doing that, see the `Accessing a service
        using SmartStack` section of http://y/smartstack.
        """
        return data_pipeline_conf.read_bool(
            'load_schematizer_host_and_port_from_smartstack',
            default=False
        )

    @property
    def schematizer_client(self):
        """Returns a bravado client for the schematizer api.

        By default, this will connect to a schematizer instance running in the
        included docker-compose file.
        """
        return SwaggerClient.from_url(
            'http://{0}/swagger.json'.format(self.schematizer_host_and_port)
        )

    @property
    def schematizer_client_max_connection_retry(self):
        """Maximum number of times schematizer_clientlib tries to connect
        to schematizer before giving up.
        """
        return data_pipeline_conf.read_float(
            'schematizer_client_max_connection_retry',
            default=5
        )

    @property
    def cluster_config(self):
        """Returns a yelp_kafka.config.ClusterConfig.

        This method will use :meth:`kafka_cluster_type` and
        :meth:`kafka_cluster_name` to fetch a `ClusterConfig` using Yelp's
        kafka discovery mechanism.  If they both aren't specified, it will
        fall back to creating a `ClusterConfig` from :meth:`kafka_broker_list`
        and :meth:`kafka_zookeeper`.  The default `ClusterConfig` will point
        at the testing docker container.
        """
        if (
            self.kafka_cluster_type is not None and
            self.kafka_cluster_name is not None and
            not self.should_use_testing_containers
        ):
            from yelp_kafka.discovery import get_kafka_cluster  # NOQA
            return get_kafka_cluster(self.kafka_cluster_type,
                                     'data_pipeline-client',
                                     self.kafka_cluster_name
                                     )
        else:
            return ClusterConfig(
                type='standard',
                name='data_pipeline',
                broker_list=self.kafka_broker_list,
                zookeeper=self.kafka_zookeeper
            )

    @property
    def kafka_cluster_type(self):
        """Cluster type corresponds to a file name in
        `/nail/etc/kafka_discovery` (ex. 'scribe' or 'standard' or 'datapipe').
        If you're setting this manually, `datapipe` is probably the cluster type
        you want.  See http://y/kafka_discovery for discovery details.
        """
        return data_pipeline_conf.read_string(
            'kafka_cluster_type',
            default=None
        )

    @property
    def kafka_cluster_name(self):
        """Cluster name corresponds to one of the named clusters in the cluster
        list in the discovery file for a cluster type.  For example,
        `/nail/etc/kafka_discovery/standard.yaml` clusters list contains a
        cluster `uswest1-devc`.  To use this cluster, :meth:`kafka_cluster_type`
        should be set to `standard`, and :meth:`kafka_cluster_name` should be
        set to `uswest1-devc`.
        """
        return data_pipeline_conf.read_string(
            'kafka_cluster_name',
            default=None
        )

    @property
    def kafka_broker_list(self):
        """:meth:`cluster_config` will use this list to construct a Kafka
        `ClusterConfig` in the absence of :meth:`kafka_cluster_type` or
        :meth:`kafka_cluster_name`.

        Each broker is expected to be a string in `host:port` form.
        """
        return data_pipeline_conf.read_list(
            'kafka_broker_list',
            default=['{0}:49255'.format(self.YOCALHOST)]
        )

    @property
    def kafka_zookeeper(self):
        """Zookeeper connection string - see kafka discovery files for addition
        detail.
        """
        return data_pipeline_conf.read_string(
            'kafka_zookeeper',
            default='{0}:32796'.format(self.YOCALHOST)
        )

    @property
    def zookeeper_discovery_path(self):
        return data_pipeline_conf.read_string(
            'zookeeper_discovery_path',
            default='zookeeper_discovery{ecosystem}.yaml'
        )

    @property
    def consumer_get_messages_timeout_default(self):
        """ Default timeout for blocking calls to ``Consumer.get_messages``
        """
        return data_pipeline_conf.read_float(
            'consumer_get_messages_timeout_default',
            default=0.1,
        )

    @property
    def consumer_partitioner_cooldown_default(self):
        """ Default partitioner cooldown time. See ``yelp_kafka.partitioner`` for
        more details.
        """
        return data_pipeline_conf.read_float(
            'consumer_partitioner_cooldown_default',
            default=0.5,
        )

    @property
    def consumer_use_group_sha_default(self):
        """Default use group sha option. See yelp_kafka.config for more details
        """
        return data_pipeline_conf.read_bool(
            'consumer_use_group_sha_default',
            default=True
        )

    @property
    def monitoring_window_in_sec(self):
        """Returns the duration(in sec) for which the monitoring system will count
        the number of messages processed by the client
        """
        return data_pipeline_conf.read_int(
            'monitoring_window_in_sec',
            default=600,
        )

    @property
    def topic_creation_wait_timeout(self):
        """Maximum time in seconds to wait for a kafka topic to be created
        during tests.
        """
        return data_pipeline_conf.read_int(
            'topic_creation_wait_timeout',
            default=60,
        )

    @property
    def skip_messages_with_pii(self):
        """Return true if we drop the message containing pii in kafka publish.
        """
        return data_pipeline_conf.read_bool(
            'skip_messages_with_pii',
            default=True
        )

    @property
    def encryption_type(self):
        """Algorithm and key to use when encrypting pii,
        e.g., 'AES_MODE_CBC-1'. The default here is None,
        but in the default data_pipeline config file,
        it is set to AES_MODE_CBC-1 for ease of set-up."""
        return data_pipeline_conf.read_string(
            'encryption_type',
            default=None
        )

    @property
    def key_location(self):
        """Directory in which to look for key to encrypt pii."""
        return data_pipeline_conf.read_string(
            'key_location',
            default='./'
        )

    @property
    def data_pipeline_teams_config_file_path(self):
        """Returns the path to the config file which specifies valid teams for
        the data pipeline.
        """
        return data_pipeline_conf.read_string(
            'data_pipeline_teams_config_file_path',
            default='teams.yaml'
        )

    @property
    def kafka_client_ack_count(self):
        """ack setting of the kafka client for publishing messages.  Default to
        -1, which indicates that the server will wait until the message is
        committed by all in-sync replicas before sending a response.  See
        https://github.com/mumrah/kafka-python/blob/master/kafka/client.py#L445
        for detail information.
        """
        return data_pipeline_conf.read_int('kafka_client_ack_count', default=-1)

    @property
    def producer_max_publish_retry_count(self):
        """Number of times the producer will retry to publish messages.
        """
        return data_pipeline_conf.read_int(
            'producer_max_publish_retry_count',
            default=5
        )

    @property
    def consumer_max_offset_retry_count(self):
        """Number of times the consumer will retry to set its offsets.
        """
        return data_pipeline_conf.read_int(
            'consumer_max_offset_retry_count',
            default=5
        )

    @property
    def kafka_producer_buffer_size(self):
        """The maximum number of messages that the clientlib will buffer
        before sending them out to kafka.
        """
        return data_pipeline_conf.read_int(
            'kafka_producer_buffer_size',
            default=5000
        )

    @property
    def kafka_producer_flush_time_limit_seconds(self):
        """The maximum amount of time in seconds that the clientlib will wait
        for its buffer to fill before forcing a flush to kafka.
        """
        return data_pipeline_conf.read_float(
            'kafka_producer_flush_time_limit_seconds',
            default=0.1
        )

    @property
    def skip_position_info_update_when_not_set(self):
        """By default, the clientlib will replace upstream position info in the
        producer with the contents from the last published message, regardless
        of content.  When this option is True, the producer will not update
        position info when it's not set in the message.

        This option is useful when a single upstream message can potentially
        result in multiple downstream messages being produced.  In that case,
        only the last message should carry upstream position information.
        """
        return data_pipeline_conf.read_bool(
            'skip_position_info_update_when_not_set',
            default=False
        )

    @property
    def merge_position_info_update(self):
        """By default, the clientlib will replace upstream position info
        returned by the producer.  When this option is set to True, the
        producer will instead merge the upstream position info dictionaries.
        When used, `upstream_position_info` should always be set, or
        :meth:`skip_position_info_update_when_not_set` should be set to True.

        This can be useful if consumption happens from multiple
        upstream sources (such as a partitioned upstream topic), and
        the producer needs to merge information for each of the upstreams.
        """
        return data_pipeline_conf.read_bool(
            'merge_position_info_update',
            default=False
        )

    @property
    def force_recovery_from_publication_unensurable_error(self):
        """Toggling this option to true and restarting the impacted service will
        force the `Producer` to recover from `PublicationUnensurableError`
        generated in the `ensure_messages_published` method.  This option should
        be turned off after the service recovers.  Ideally it would never be
        used, but it can be helpful to get an exactly-once service working again
        after a logical error, at the expense of violating the guarantee.  It's
        primarily here to get services running again, while we're working out
        their kinks and debugging.  Don't use it unless you're absolutely sure
        of what you're doing.

        Warning:
            Setting this option to True guarantees that the exactly-once
            guarantee will be broken.
        """
        return data_pipeline_conf.read_bool(
            'force_recovery_from_publication_unensurable_error',
            default=False
        )

    @property
    def sensu_ping_window_seconds(self):
        """The ping window defines the minimum time (in seconds) the producer will wait
        prior to sending another OK message to sensu.  For example, say the ping window
        is 30 seconds, and an event is published at time=0.  No events prior to time=30s
        will cause the producer to publish an OK message to sensu.  Once an event comes
        in after time=30s the clock will be reset.  That is, if the next event comes in
        at time=40s the producer will send the OK, and not publish OK's again until
        after time=70s.  The purpose of this is to throtte the number events each
        producer sends to sensu.
        """
        return data_pipeline_conf.read_int('sensu_ping_window_seconds', default=30)

    @property
    def expected_heartbeat_interval(self):
        """The producer expects the upstream to send heartbeats at intervels with a maximum
        of this number of seconds.  For example, if this returns 300 then we expect the
        upstream to create at least one event every 300 seconds.
        """
        return data_pipeline_conf.read_int('expected_heartbeat_interval', default=300)

    @property
    def sensu_ttl(self):
        """This is the time to live (TTL) for the producer.  If sensu doesn't get an OK
        message by the time to live interval sensu will alert.  Note that the TTL is derived
        from the expected heartbeat and the ping window to prevent false alarms.  For example
        if the ping window is 30 seconds and the hearbeat interval is 300 seconds the TTL
        will be 331 seconds, which guarantees if things are working correctly we'll have be
        sending OK's to sensu on time.  If either the producer goes down or the upstream
        stops sending heartbeats to the producer the team owning the producer will be alerted
        within one TTL period.
        """
        return self.expected_heartbeat_interval + self.sensu_ping_window_seconds + 1

    @property
    def sensu_host(self):
        """If we're running in Paasta, use the paasta cluster from the
        environment directly as laid out in PAASTA-1579.  This makes it so that
        local-run and real sensu alerts go to the same cluster, which should
        prevent false alerts that never resolve when we run locally.
        """
        if os.environ.get('PAASTA_CLUSTER'):
            return "paasta-{cluster}.yelp".format(
                cluster=os.environ.get('PAASTA_CLUSTER')
            )
        else:
            return data_pipeline_conf.read_string('sensu_host', self.YOCALHOST)

    @property
    def sensu_page_on_critical(self):
        """If this is true the client will page on a sensu critical alert"""
        return data_pipeline_conf.read_bool('sensu_page_on_critical', default=False)

    @property
    def container_name(self):
        """This is the name of the paasta instance in which the client is running.
        For example: canary"""
        return os.environ.get(
            'PAASTA_INSTANCE',
            data_pipeline_conf.read_string('container_name', "no_paasta_container")
        )

    @property
    def container_env(self):
        """This is the name of the paasta cluster in which the client is running.
        For example: norcal-prod.
        """
        return os.environ.get(
            'PAASTA_CLUSTER',
            data_pipeline_conf.read_string('container_env', "no_paasta_environment")
        )

    @property
    def sensu_source(self):
        """This ensures that the alert tracks both the paasta environment and
        the running instance, so we can have separate alerts for the canary
        and the main instances.
        """
        return '{container_env}_{container_name}'.format(
            container_env=self.container_env,
            container_name=self.container_name
        )

    @property
    def sensu_alert_after_seconds(self):
        """This ensures that the alert tracks both the paasta environment and
        the running instance, so we can have separate alerts for the canary
        and the main instances.
        """
        return "{}s".format(
            data_pipeline_conf.read_int(
                'sensu_alert_after_seconds',
                default=300
            )
        )

    @property
    def enable_sensu(self):
        """enable sensu alerting"""
        return data_pipeline_conf.read_bool('enable_sensu', False)

    @property
    def enable_meteorite(self):
        """enable meteorite monitoring"""
        return data_pipeline_conf.read_bool('enable_meteorite', False)

    @property
    def max_producer_delay_seconds(self):
        """This is the maximum number of seconds allowed between the event times
        and the time the producer is to publish the message to kafka.  Anything
        greater than this time will produce an alert"""
        return data_pipeline_conf.read_int('sensu_max_delay_seconds', 300)

    @property
    def ecosystem_file_patch(self):
        return data_pipeline_conf.read_string('ecosystem_file_patch', "/nail/etc/ecosystem")


def configure_from_dict(config_dict):
    """Configure the :mod:`data_pipeline` clientlib from a dictionary.

    Args:
        config_dict (dict): a dict of config data
    """
    staticconf.DictConfiguration(config_dict, namespace=namespace)


_config = Config()


def get_config():
    """Returns the global data pipeline configuration object"""
    return _config



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import namedtuple

from data_pipeline.schematizer_clientlib.models.data_target import _DataTarget
from data_pipeline.schematizer_clientlib.models.model_base import BaseModel


"""
Represent the data of a consumer group.  A consumer group represents a group of
the consumers that send all the messages to the same destination, defined as a
"data target".

Args:
    consumer_group_id (int): The id of the consumer group.
    group_name (str): The name of the consumer group.
    data_target (data_pipeline.schematizer_clientlib.models.data_target.DataTarget):
        The data_target this consumer group associates to.
"""
ConsumerGroup = namedtuple(
    'ConsumerGroup',
    ['consumer_group_id', 'group_name', 'data_target']
)


class _ConsumerGroup(BaseModel):
    """Internal class used to convert from/to various data structure and
    facilitate constructing the return value of schematizer functions.
    """

    def __init__(self, consumer_group_id, group_name, data_target):
        self.consumer_group_id = consumer_group_id
        self.group_name = group_name
        self.data_target = data_target

    @classmethod
    def from_response(cls, response):
        return cls(
            consumer_group_id=response.consumer_group_id,
            group_name=response.group_name,
            data_target=_DataTarget.from_response(response.data_target)
        )

    def to_cache_value(self):
        return {
            'consumer_group_id': self.consumer_group_id,
            'group_name': self.group_name,
            'data_target_id': self.data_target.data_target_id
        }

    @classmethod
    def from_cache_value(cls, cache_value):
        id_only_data_target = _DataTarget(
            data_target_id=cache_value['data_target_id'],
            name=None,
            target_type=None,
            destination=None
        )
        return cls(
            consumer_group_id=cache_value['consumer_group_id'],
            group_name=cache_value['group_name'],
            data_target=id_only_data_target
        )

    def to_result(self):
        return ConsumerGroup(
            consumer_group_id=self.consumer_group_id,
            group_name=self.group_name,
            data_target=self.data_target.to_result()
        )



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import time
from collections import defaultdict
from collections import namedtuple
from contextlib import contextmanager

from cached_property import cached_property
from kafka import create_message
from kafka import KafkaClient
from kafka.common import ProduceRequest

from data_pipeline._position_data_tracker import PositionDataTracker
from data_pipeline._producer_retry import RetryHandler
from data_pipeline._retry_util import ExpBackoffPolicy
from data_pipeline._retry_util import MaxRetryError
from data_pipeline._retry_util import Predicate
from data_pipeline._retry_util import retry_on_condition
from data_pipeline._retry_util import RetryPolicy
from data_pipeline.config import get_config
from data_pipeline.envelope import Envelope


_EnvelopeAndMessage = namedtuple("_EnvelopeAndMessage", ["envelope", "message"])
logger = get_config().logger


# prepare needs to be in the module top level so it can be serialized for
# multiprocessing
def _prepare(envelope_and_message):
    try:
        kwargs = {}
        if envelope_and_message.message.keys:
            kwargs['key'] = envelope_and_message.message.encoded_keys
        return create_message(
            envelope_and_message.envelope.pack(envelope_and_message.message),
            **kwargs
        )
    except:
        logger.exception('Prepare failed')
        raise


class KafkaProducer(object):
    """The KafkaProducer deals with buffering messages that need to be published
    into Kafka, preparing them for publication, and ultimately publishing them.

    Args:
        producer_position_callback (function): The producer position callback
            is called when the KafkaProducer is instantiated, and every time
            messages are published to notify the producer of current position
            information of successfully published messages.
        dry_run (Optional[bool]): When dry_run mode is on, the producer won't
            talk to real KafKa topic, nor to real Schematizer.  Default to False.
    """
    @cached_property
    def envelope(self):
        return Envelope()

    def __init__(self, producer_position_callback, dry_run=False):
        self.producer_position_callback = producer_position_callback
        self.dry_run = dry_run
        self.kafka_client = KafkaClient(get_config().cluster_config.broker_list)
        self.position_data_tracker = PositionDataTracker()
        self._reset_message_buffer()
        self.skip_messages_with_pii = get_config().skip_messages_with_pii
        self._publish_retry_policy = RetryPolicy(
            ExpBackoffPolicy(with_jitter=True),
            max_retry_count=get_config().producer_max_publish_retry_count
        )
        self._automatic_flush_enabled = True

    @contextmanager
    def disable_automatic_flushing(self):
        """Prevents the producer from flushing automatically (e.g. for timeouts
        or batch size) while the context manager is open.
        """
        try:
            self._automatic_flush_enabled = False
            yield
        finally:
            self._automatic_flush_enabled = True

    def wake(self):
        """Should be called periodically if we're not otherwise waking up by
        publishing, to ensure that messages are actually published.
        """
        # if we haven't woken up in a while, we may need to flush messages
        self._flush_if_necessary()

    def publish(self, message):
        if message.contains_pii and self.skip_messages_with_pii:
            logger.info(
                "Skipping a PII message - "
                "uuid hex: {0}, "
                "schema_id: {1}, "
                "timestamp: {2}, "
                "type: {3}".format(
                    message.uuid_hex,
                    message.schema_id,
                    message.timestamp,
                    message.message_type.name
                )
            )
            return
        self._add_message_to_buffer(message)
        self.position_data_tracker.record_message_buffered(message)
        self._flush_if_necessary()

    def flush_buffered_messages(self):
        produce_method = (self._publish_produce_requests_dry_run
                          if self.dry_run else self._publish_produce_requests)
        produce_method(self._generate_produce_requests())
        self._reset_message_buffer()

    def close(self):
        self.flush_buffered_messages()
        self.kafka_client.close()

    def _publish_produce_requests(self, requests):
        """It will try to publish all the produce requests for topics, and
        retry a number of times until either all the requests are successfully
        published or it can no longer retry, in which case, the exception will
        be thrown.

        Each time the requests that are successfully published in the previous
        round will be removed from the requests and won't be published again.
        """
        unpublished_requests = list(requests)
        retry_handler = RetryHandler(self.kafka_client, unpublished_requests)

        def has_requests_to_be_sent():
            return bool(retry_handler.requests_to_be_sent)

        retry_handler = retry_on_condition(
            retry_policy=self._publish_retry_policy,
            retry_conditions=[Predicate(has_requests_to_be_sent)],
            func_to_retry=self._publish_requests,
            use_previous_result_as_param=True,
            retry_handler=retry_handler
        )
        if retry_handler.has_unpublished_request:
            raise MaxRetryError(last_result=retry_handler)

    def _publish_requests(self, retry_handler):
        """Main function to publish message requests.  This function is wrapped
        with retry function and will be retried based on specified retry policy

        Args:
            retry_handler: :class:`data_pipeline._producer_retry.RetryHandler`
                that determines which messages should be retried next time.
        """
        if not retry_handler.requests_to_be_sent:
            return retry_handler

        responses = self._try_send_produce_requests(
            retry_handler.requests_to_be_sent
        )

        retry_handler.update_requests_to_be_sent(
            responses,
            self.position_data_tracker.topic_to_kafka_offset_map
        )
        self._record_success_requests(retry_handler.success_topic_stats_map)
        return retry_handler

    def _try_send_produce_requests(self, requests):
        # Either it throws exceptions and none of them succeeds, or it returns
        # responses of all the requests (success or fail response).
        try:
            return self.kafka_client.send_produce_request(
                payloads=requests,
                acks=get_config().kafka_client_ack_count,
                fail_on_error=False
            )
        except Exception:
            # Exceptions like KafkaUnavailableError, LeaderNotAvailableError,
            # UnknownTopicOrPartitionError, etc., are not controlled by
            # `fail_on_error` flag and could be thrown from the kafka client,
            # and fail all the requests.  We will retry all the requests until
            # either all of them are successfully published or it exceeds the
            # maximum retry criteria.
            return []

    def _record_success_requests(self, success_topic_stats_map):
        for topic_partition, stats in success_topic_stats_map.iteritems():
            topic = topic_partition.topic_name
            assert stats.message_count == len(self.message_buffer[topic])
            self.position_data_tracker.record_messages_published(
                topic=topic,
                offset=stats.original_offset,
                message_count=stats.message_count
            )
            self.message_buffer.pop(topic)

    def _publish_produce_requests_dry_run(self, requests):
        for request in requests:
            self._publish_single_request_dry_run(request)

    def _publish_single_request_dry_run(self, request):
        topic = request.topic
        message_count = len(request.messages)
        self.position_data_tracker.record_messages_published(
            topic,
            -1,
            message_count
        )

    def _is_ready_to_flush(self):
        time_limit = get_config().kafka_producer_flush_time_limit_seconds
        return (self._automatic_flush_enabled and (
            (time.time() - self.start_time) >= time_limit or
            self.message_buffer_size >= get_config().kafka_producer_buffer_size
        ))

    def _flush_if_necessary(self):
        if self._is_ready_to_flush():
            self.flush_buffered_messages()

    def _add_message_to_buffer(self, message):
        topic = message.topic
        message = self._prepare_message(message)

        self.message_buffer[topic].append(message)
        self.message_buffer_size += 1

    def _generate_produce_requests(self):
        return [
            ProduceRequest(topic=topic, partition=0, messages=messages)
            for topic, messages in self._generate_prepared_topic_and_messages()
        ]

    def _generate_prepared_topic_and_messages(self):
        return self.message_buffer.iteritems()

    def _prepare_message(self, message):
        return _prepare(_EnvelopeAndMessage(envelope=self.envelope, message=message))

    def _reset_message_buffer(self):
        if not hasattr(self, 'message_buffer_size') or self.message_buffer_size > 0:
            self.producer_position_callback(self.position_data_tracker.get_position_data())
        self.start_time = time.time()
        self.message_buffer = defaultdict(list)
        self.message_buffer_size = 0


class LoggingKafkaProducer(KafkaProducer):
    def _publish_produce_requests(self, requests):
        logger.info(
            "Flushing buffered messages - requests={0}, messages={1}".format(
                len(requests), self.message_buffer_size
            )
        )
        try:
            super(LoggingKafkaProducer, self)._publish_produce_requests(requests)
            logger.info("All messages published successfully")
        except MaxRetryError as e:
            logger.exception(
                "Failed to publish all produce requests. {0}".format(repr(e))
            )
            raise

    def _reset_message_buffer(self):
        logger.info("Resetting message buffer for success requests.")
        super(LoggingKafkaProducer, self)._reset_message_buffer()

    def _publish_single_request_dry_run(self, request):
        super(LoggingKafkaProducer, self)._publish_single_request_dry_run(request)
        logger.debug("dry_run mode: Would have published {0} messages to {1}".format(
            len(request.messages),
            request.topic
        ))



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import datetime
import errno
import fileinput
import gzip
import json
import os


class CompressedStreamRotator(object):
    """Reads from standard in, compresses the input and writes it to files,
    rotating the files periodically.  This is designed to take streaming input
    and convert it to daily files that can be uploaded to s3 and processed with
    Mycroft or other log tools.

    File status information is output on standard out.
    """

    def __init__(self):
        self.file_start_time = None
        self.output_file = None

    def run(self):
        try:
            self._compress_streaming_json()
        except IOError as e:
            if e.errno == errno.EPIPE:
                # just stop if the pipe breaks
                pass
            else:
                raise
        finally:
            self._close_file()

    def _compress_streaming_json(self):
        for line in fileinput.input():
            line = line.strip()
            self._process_line(line)

    def _process_line(self, line):
        self._rotate_to_next_output_file(json.loads(line)['timestamp'])
        self.output_file.write(line)
        self.output_file.write("\n")

    def _rotate_to_next_output_file(self, timestamp):
        current_time = datetime.datetime.fromtimestamp(timestamp)
        if not self._belongs_in_current_file(current_time):
            print current_time, self.file_start_time
            self._close_file()
            self._create_complete_file_if_needed(current_time)
            self._open_file(current_time)

    def _belongs_in_current_file(self, current_time):
        return (
            self.file_start_time is not None and
            current_time < self._get_file_end_time() and
            self._file_date_matches_current_time(current_time)
        )

    def _file_date_matches_current_time(self, current_time):
        return self.file_start_time is not None and self.file_start_time.date() >= current_time.date()

    def _create_complete_file_if_needed(self, current_time):
        if self.file_start_time is not None and not self._file_date_matches_current_time(current_time):
            with open(os.path.join(self._get_current_file_path(), 'COMPLETE'), 'w'):
                # Just create the file
                pass

    def _get_current_file_path(self):
        return os.path.abspath(self.file_start_time.date().strftime("%Y/%m/%d"))

    def _get_file_end_time(self):
        return self.file_start_time + datetime.timedelta(minutes=15)

    def _ensure_current_path_exists(self):
        if not os.path.exists(self._get_current_file_path()):
            os.makedirs(self._get_current_file_path())

    def _open_file(self, start_time):
        self.file_start_time = start_time
        self._ensure_current_path_exists()

        file_name = "%s.json.gz" % start_time.isoformat()
        file_path = os.path.join(self._get_current_file_path(), file_name)
        self.output_file = gzip.open(file_path, 'wb', 3)

    def _close_file(self):
        if self.output_file is not None:
            self.output_file.close()
        self.output_file = None


if __name__ == "__main__":
    CompressedStreamRotator().run()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import defaultdict
from contextlib import contextmanager

from cached_property import cached_property
from kafka import KafkaClient
from kafka.common import FailedPayloadsError
from kafka.common import OffsetCommitRequest
from kafka.util import kafka_bytestring
from yelp_kafka import discovery
from yelp_kafka.config import KafkaConsumerConfig

from data_pipeline._consumer_tick import _ConsumerTick
from data_pipeline._retry_util import ExpBackoffPolicy
from data_pipeline._retry_util import retry_on_exception
from data_pipeline._retry_util import RetryPolicy
from data_pipeline.client import Client
from data_pipeline.config import get_config
from data_pipeline.consumer_source import FixedSchemas
from data_pipeline.envelope import Envelope
from data_pipeline.message import Message
from data_pipeline.schematizer_clientlib.schematizer import get_schematizer

logger = get_config().logger


class MultipleClusterTypeError(Exception):
    def __init__(self, *cluster_types):
        err_message = (
            "Consumer can not process topics from different kafka cluster "
            "types, i.e. (" + ", ".join(cluster_types) + ")."
        )
        Exception.__init__(self, err_message)


class TopicNotFoundInRegionError(Exception):
    def __init__(self, topic_name, cluster_type, cluster_name):
        err_message = (
            "Topic `{}` not found in `{}` kafka type cluster in `{}` "
            "region".format(topic_name, cluster_type, cluster_name)
        )
        Exception.__init__(self, err_message)


class ConsumerTopicState(object):
    """
    ConsumerTopicState object holds the state of a consumer topic with
    partition_offset_map mapping all the partitions of the topic with their
    last seen offsets and last_seen_schema_id with the schema id of the
    latest processed message.
    Whenever consumer receives a yelp_kafka message for a topic, it builds a
    data_pipeline.message.Message and updates the topic state map for the
    topic with updated ConsumerTopicState (updates last_seen_schema_id with
    the schema_id of the message and partition offset with the offset of
    the message).

    Args:
        partition_offset_map ({int:int}): map of partitions to their last
            seen offsets.
        last_seen_schema_id: The last seen schema_id.
    """

    def __init__(self, partition_offset_map, last_seen_schema_id):
        self.partition_offset_map = partition_offset_map
        self.last_seen_schema_id = last_seen_schema_id

    def __repr__(self):
        return '<ConsumerTopicState(last_seen_schema_id: {0}, partition_offset_map: {1})>'.format(
            self.last_seen_schema_id,
            self.partition_offset_map
        )


class BaseConsumer(Client):
    """
    This is the base interface for the Consumer implementations. It provides
    a set of methods common to the derived consumer classes and specifies
    which methods should be overridden. It should not be instantiated and
    has only been made public to ensure docs are generated.

    Args:
        consumer_name (str): See parameter `client_name` in
            :class:`data_pipeline.client.Client`.  The `consumer_name` will
            be registered with Kafka to commit offsets.
        team_name (str): See parameter `team_name` in
            :class:`data_pipeline.client.Client`.
        expected_frequency_seconds (int, ExpectedFrequency): See parameter
            `expected_frequency_seconds` in :class:`data_pipeline.client.Client`.
        topic_to_consumer_topic_state_map ({str:Optional(ConsumerTopicState)}):
            A map of topic names to `ConsumerTopicState` objects which define
            the offsets to start from. The ConsumerTopicState of a topic may be
            `None`, in which case the committed kafka offset for the
            consumer_name is used. If there is no committed kafka offset for
            the consumer_name the consumer will begin from the
            `auto_offset_reset` offset in the topic.
        consumer_source (ConsumerSource): Object to specify the topics this
            consumer consumes messages from. It must be a
            :class:`data_pipeline.consumer_source.ConsumerSource` object. For
            example, to process messages from a fixed set of topics, use
            :class:`data_pipeline.consumer_source.FixedTopics`.
            In case of FixedSchema consumer source, at most one schema_id can
            be provided per topic. Consumer would use that schema as reader
            schema to decode the message. In case no schema id for a topic is
            specified, then the Consumer would use the schema id that the
            message was encoded with to decode the message.
        auto_offset_reset (str): automatically resets the offset when there is
            no initial offset in Zookeeper or if an offset is out of range.
            If 'largest', reset the offset to the latest available message (tail).
            If 'smallest' reset from the earliest (head).
        partitioner_cooldown (float): Waiting time (in seconds) for the
            consumer to acquire the partitions. See
            yelp_kafka/yelp_kafka/partitioner.py for more details
        use_group_sha (Optional[boolean]): Used by partitioner to establish
            group membership. If false, consumer group with same name will
            be treated as the same group; otherwise, they will be different
            since group sha is different. Default is true.
        pre_rebalance_callback (Optional[Callable[{str:list[int]}, None]]):
            Optional callback which is passed a dict of topic as key and list
            of partitions as value. It's important to note this may be called
            multiple times in a single repartition, so any actions taken as a
            result must be idempotent. You are guaranteed that no messages will
            be consumed between this callback and the post_rebalance_callback.
        post_rebalance_callback (Optional[Callable[{str:list[int]}, None]]):
            Optional callback which is passed a dict of topic as key and list
            of partitions as value which were acquired in a repartition. You
            are guaranteed that no messages will be consumed between the
            pre_rebalance_callback and this callback.
        fetch_offsets_for_topics: (Optional[Callable[List[str],
            Dict[str, Optional[Dict[int, int]]]]]): Optional callback which is
            passed a list of topics, and should return a dictionary where keys
            are topic names and values are either None if no offset should be
            manually set, or a map from partition to offset.
            If implemented, this function will be called every time
            consumer refreshes the topics, so that the consumer can provide a
            map of partitions to offsets for each topic, or None if the default
            behavior should be employed instead. The default behavior is
            picking up the last committed offsets of topics.
            This method must be implemented if topic state is to be stored
            in some system other than Kafka, for example when writing data from
            Kafka into a transactional store.
        pre_topic_refresh_callback: (Optional[Callable[[set[str], set[str]],
            Any]]): Optional callback that gets executed right before the
            consumer is about to refresh the topics. The callback function is
            passed in a set of topic names Consumer is currently consuming
            from (current_topics) and a set of topic names Consumer will be
            consuming from (refreshed_topics). The return value of the
            function is ignored.
        cluster_name: (Optional[string]): Indicates the name (region) of kafka
            cluster the topics belong to. Ex: uswest2-prod. Accordingly the
            Consumer will connect to Kafka cluster in the corresponding region.
            All topics should belong to the same kafka cluster name.
            Defaults to None.
    """

    def __init__(
        self,
        consumer_name,
        team_name,
        expected_frequency_seconds,
        topic_to_consumer_topic_state_map=None,
        consumer_source=None,
        force_payload_decode=True,
        auto_offset_reset='smallest',
        partitioner_cooldown=get_config().consumer_partitioner_cooldown_default,
        use_group_sha=get_config().consumer_use_group_sha_default,
        topic_refresh_frequency_seconds=get_config().topic_refresh_frequency_seconds,
        pre_rebalance_callback=None,
        post_rebalance_callback=None,
        fetch_offsets_for_topics=None,
        pre_topic_refresh_callback=None,
        cluster_name=None
    ):
        super(BaseConsumer, self).__init__(
            consumer_name,
            team_name,
            expected_frequency_seconds,
            monitoring_enabled=False
        )

        if ((topic_to_consumer_topic_state_map and consumer_source) or
                (not topic_to_consumer_topic_state_map and not consumer_source)):
            raise ValueError("Exactly one of topic_to_consumer_topic_state_map "
                             "or consumer_source must be specified")

        self.consumer_source = consumer_source
        self.topic_to_consumer_topic_state_map = topic_to_consumer_topic_state_map
        self.force_payload_decode = force_payload_decode
        self.auto_offset_reset = auto_offset_reset
        self.partitioner_cooldown = partitioner_cooldown
        self.use_group_sha = use_group_sha
        self.running = False
        self.consumer_group = None
        self.pre_rebalance_callback = pre_rebalance_callback
        self.post_rebalance_callback = post_rebalance_callback
        self.fetch_offsets_for_topics = fetch_offsets_for_topics
        self.pre_topic_refresh_callback = pre_topic_refresh_callback
        self.cluster_name = self._set_cluster_name(cluster_name)
        self._refresh_timer = _ConsumerTick(
            refresh_time_seconds=topic_refresh_frequency_seconds
        )
        self._topic_to_reader_schema_map = self._get_topic_to_reader_schema_map(
            consumer_source
        )
        self._consumer_retry_policy = RetryPolicy(
            ExpBackoffPolicy(with_jitter=True),
            max_retry_count=get_config().consumer_max_offset_retry_count
        )
        self._envelope = Envelope()
        if self.topic_to_consumer_topic_state_map:
            self.cluster_type = self._determine_cluster_type_from_topics(
                self.topic_to_consumer_topic_state_map.keys()
            )

    @cached_property
    def _schematizer(self):
        return get_schematizer()

    def _set_cluster_name(self, cluster_name):
        return cluster_name or get_config().kafka_cluster_name

    def _determine_cluster_type_from_topics(self, topic_names):
        """Checks whether the cluster type of all topics are the same and
        returns it.

        :param topic_names: list of topic names
        :raises MultipleClusterTypeError: if all topics don't have the same cluster type.
        :return: cluster_type
        """
        cluster_type = None
        for topic_name in topic_names:
            topic = self._schematizer.get_topic_by_name(topic_name)
            if cluster_type is None:
                cluster_type = topic.cluster_type
            elif cluster_type != topic.cluster_type:
                raise MultipleClusterTypeError(
                    cluster_type,
                    topic.cluster_type
                )
        return cluster_type

    def _get_refreshed_topic_to_consumer_topic_state_map(
        self,
        topic_to_consumer_state_map,
        consumer_source
    ):
        if topic_to_consumer_state_map:
            return topic_to_consumer_state_map
        if consumer_source:
            return self._get_topic_to_offset_map(consumer_source.get_topics())
        return {}

    def _get_topic_to_offset_map(self, topics):
        """ This function constructs a new topic_to_consumer_topic_state_map
        from the given topics set. Return the topic_to_consumer_topic_state_map
        dictionary.
        """
        if self.fetch_offsets_for_topics:
            topic_to_partition_offsets_map = self.fetch_offsets_for_topics(
                list(topics)
            )
            return {
                topic: (
                    ConsumerTopicState(
                        partition_offset_map,
                        last_seen_schema_id=None
                    ) if partition_offset_map else None
                )
                for topic, partition_offset_map in
                topic_to_partition_offsets_map.iteritems()
            }
        else:
            return {topic: None for topic in topics}

    def _get_topic_to_reader_schema_map(self, consumer_source):
        """
        In case of FixedSchema consumer source, creates a dictionary with topic
        name as key and reader schema id as value that would be used to decode
        the messages of that topic. Return the topic_to_reader_schema_map
        dictionary.
        """
        topic_to_reader_schema_map = {}
        if isinstance(consumer_source, FixedSchemas):
            schema_to_topic_map = consumer_source.get_schema_to_topic_map()
            for schema_id, topic_name in schema_to_topic_map.iteritems():
                if topic_name not in topic_to_reader_schema_map:
                    topic_to_reader_schema_map[topic_name] = schema_id
                else:
                    raise ValueError(
                        'Multiple reader schemas ({schema1}, {schema2}) exist '
                        'for topic: {topic}'.format(
                            schema1=topic_to_reader_schema_map[topic_name],
                            schema2=schema_id,
                            topic=topic_name
                        )
                    )
        return topic_to_reader_schema_map

    def _set_topic_to_partition_map(self, topic_to_consumer_topic_state_map):
        """ This function takes a topic_to_consumer_topic_state_map and sets
        the topic_to_partition_map instance variable with topic names as keys
        and corresponding partition lists as values.
        """
        self.cluster_type = self._determine_cluster_type_from_topics(
            topic_to_consumer_topic_state_map.keys()
        )
        self.topic_to_partition_map = {}
        for (
            topic_name, consumer_topic_state
        ) in topic_to_consumer_topic_state_map.iteritems():
            topics = self._get_topics_in_region_from_topic_name(topic_name)
            for topic in topics:
                self.topic_to_partition_map[topic] = (
                    consumer_topic_state.partition_offset_map.keys()
                    if consumer_topic_state else None
                )
        self._set_registrar_tracked_schema_ids(topic_to_consumer_topic_state_map)

    def _get_topics_in_region_from_topic_name(self, topic_name):
        """We use yelp_kafka discovery to get topic for a specific
        cluster_type and cluster_name. There should 0 or 1 topic for a given
        name in a particular cluster_type and cluster_name. However the
        discovery methods returns a list of topics. Hence keeping yelp_kafka
        as the source of truth for topic info, we will tail all topics that it
        returns, which almost always is 0 or 1.
        """
        if self.cluster_type == 'scribe':
            return self._get_scribe_topics_from_topic_name(topic_name)
        else:
            return self._get_kafka_topics_from_topic_name(topic_name)

    def _get_scribe_topics_from_topic_name(self, topic_name):
        """yelp_kafka.discovery.get_region_logs_stream returns a list of
        ([topics], cluster) tuple. However if the region is specified we should
        have 0 or 1 tuple of ([topics], cluster) returned
        http://servicedocs.yelpcorp.com/docs/yelp_kafka/discovery.html#yelp_kafka.discovery.get_region_logs_stream
        """
        topics_in_region = discovery.get_region_logs_stream(
            client_id=self.client_name,
            stream=topic_name,
            region=self.cluster_name
        )
        if not topics_in_region:
            raise TopicNotFoundInRegionError(
                topic_name,
                self.cluster_type,
                self.cluster_name
            )
        topics, _ = topics_in_region[0]
        return topics

    def _get_kafka_topics_from_topic_name(self, topic_name):
        """yelp_kafka.discovery.search_topic returns a list of (topic, cluster)
        tuple. However if the region is specified we should have 0 or 1 tuple
        of (topic, cluster) returned
        http://servicedocs.yelpcorp.com/docs/yelp_kafka/discovery.html#yelp_kafka.discovery.search_topic
        """
        topics_in_region = discovery.search_topic(
            topic=topic_name,
            clusters=[self._region_cluster_config]
        )
        if not topics_in_region:
            raise TopicNotFoundInRegionError(
                topic_name,
                self.cluster_type,
                self._region_cluster_config.name
            )
        return [topic for topic, _ in topics_in_region]

    def _set_registrar_tracked_schema_ids(self, topic_to_consumer_topic_state_map):
        """
        Register what schema ids to track based on the topic map passed in
        on initialization and during topic change.
        """
        schema_id_list = [
            consumer_topic_state.last_seen_schema_id
            for consumer_topic_state in topic_to_consumer_topic_state_map.itervalues()
            if consumer_topic_state and consumer_topic_state.last_seen_schema_id
        ]
        self.registrar.register_tracked_schema_ids(schema_id_list)

    @cached_property
    def kafka_client(self):
        """ Returns the `KafkaClient` object."""
        return KafkaClient(self._region_cluster_config.broker_list)

    @property
    def client_type(self):
        """ Returns a string identifying the client type. """
        return "consumer"

    def __enter__(self):
        self.start()
        self.registrar.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            self.stop()
        except:
            logger.exception("Failed to stop the Consumer.")
            if exc_type is None:
                # The exception shouldn't mask the original exception if there
                # is one, but if an exception occurs, we want it to show up.
                raise
        # Returning any kind of truthy value will suppress the exception, if
        # there was one.  The intention of returning False here is to never
        # suppress the exception.  See:
        # https://docs.python.org/2/reference/datamodel.html#object.__exit__
        # for more information.
        return False

    def start(self):
        """ Start the Consumer. Normally this should NOT be called directly,
        rather the Consumer should be used as a context manager, which will
        start automatically when the context enters.

        Note:
            The derived class must implement _start().
        """
        self.reset_topic_to_partition_offset_cache()
        refreshed_topic_to_consumer_topic_state_map = (
            self._get_refreshed_topic_to_consumer_topic_state_map(
                self.topic_to_consumer_topic_state_map,
                self.consumer_source
            ))
        if refreshed_topic_to_consumer_topic_state_map:
            self._commit_topic_offsets(
                refreshed_topic_to_consumer_topic_state_map
            )
            # After committing partition offsets, the
            # self.topic_to_consumer_topic_state_map is set to `None` to
            # prevent committing these offsets again.
            self.topic_to_consumer_topic_state_map = None
            self._set_topic_to_partition_map(
                refreshed_topic_to_consumer_topic_state_map
            )
        self._start_consumer()

    def _start_consumer(self):
        logger.info("Starting Consumer '{0}'...".format(self.client_name))
        if self.running:
            raise RuntimeError("Consumer '{0}' is already running".format(
                self.client_name
            ))
        self._start()
        self.running = True
        logger.info("Consumer '{0}' started".format(self.client_name))

    def stop(self):
        """ Stop the Consumer. Normally this should NOT be called directly,
        rather the Consumer should be used as a context manager, which will
        stop automatically when the context exits.

        Note:
            The derived class must implement _stop().
        """
        logger.info("Stopping Consumer '{0}'...".format(self.client_name))
        if self.running:
            self._stop()
        self.registrar.stop()
        self.kafka_client.close()
        self.reset_topic_to_partition_offset_cache()
        self.running = False
        logger.info("Consumer '{0}' stopped".format(self.client_name))

    def __iter__(self):
        while True:
            yield self.get_message(
                blocking=True,
                timeout=get_config().consumer_get_messages_timeout_default
            )

    def reset_topic_to_partition_offset_cache(self):
        self.topic_to_partition_offset_map_cache = defaultdict(dict)

    def get_message(
        self,
        blocking=False,
        timeout=get_config().consumer_get_messages_timeout_default
    ):
        """ Retrieve a single message. Returns None if no message could
        be retrieved within the timeout.

        Warning:
            If `blocking` is True and `timeout` is None this will block until
            a message is retrieved, potentially blocking forever. Please be
            absolutely sure this is what you are intending if you use these
            options!

        Args:
            blocking (boolean): Set to True to block while waiting for messages
                if the buffer has been depleted. Otherwise returns immediately
                if the buffer reaches depletion.
            timeout (double): Maximum time (in seconds) to wait if blocking is
                set to True. Set to None to wait indefinitely.

        Returns:
            (Optional(data_pipeline.message.Message)): Message object or None
            if no message could be retrieved.
        """
        return next(iter(
            self.get_messages(
                count=1,
                blocking=blocking,
                timeout=timeout
            )),
            None
        )

    def get_messages(
        self,
        count,
        blocking=False,
        timeout=get_config().consumer_get_messages_timeout_default
    ):
        """ Retrieve a list of messages from the message buffer, optionally
        blocking until the requested number of messages has been retrieved.

        Note:
            The derived class must implement this method.

        Warning:
            If `blocking` is True and `timeout` is None this will block until
            the requested number of messages is retrieved, potentially blocking
            forever. Please be absolutely sure this is what you are intending
            if you use these options!

        Args:
            count (int): Number of messages to retrieve
            blocking (boolean): Set to True to block while waiting for messages
                if the buffer has been depleted. Otherwise returns immediately
                if the buffer reaches depletion.
            timeout (double): Maximum time (in seconds) to wait if blocking is
                set to True. Set to None to wait indefinitely.

        Returns:
            ([data_pipeline.message.Message]): List of Message objects with a
                of maximum size `count`, but may be smaller or empty depending
                on how many messages were retrieved within the timeout.
        """
        raise NotImplementedError

    def commit_message(self, message):
        """ Commit the offset information of a message to Kafka. Until a message
        is committed the stored kafka offset for this `consumer_name` is not updated.

        Recommended to avoid calling this too frequently, as it is relatively
        expensive.

        Args:
            message (data_pipeline.message.Message): The message to commit
        """
        return self.commit_messages([message])

    def commit_messages(self, messages):
        """ Commit the offset information of a list of messages to Kafka. Until
        a message is committed the stored kafka offset for this `consumer_name`
        is not updated.

        Recommended to avoid calling this too frequently, as it is relatively
        expensive.

        If more than one of the given messages share the same `topic/partition`
        combination then the message with the highest offset is committed.

        Note:
            It's recommended to retrieve messages in batches via
            `get_messages(..)`, do your work with them, and then commit them as
            a group with a single call to `commit_messages(..)`

        Args:
            messages [data_pipeline.message.Message]: List of messages to commit
        """
        topic_to_partition_offset_map = {}
        for message in messages:
            pos_info = message.kafka_position_info
            partition_offset_map = topic_to_partition_offset_map.get(
                message.topic,
                {}
            )
            max_offset = partition_offset_map.get(pos_info.partition, 0)
            # Increment the offset value by 1 so the consumer knows where to
            # retrieve the next message.
            partition_offset_map[pos_info.partition] = (
                max(pos_info.offset, max_offset) + 1
            )
            topic_to_partition_offset_map[message.topic] = partition_offset_map
        self.commit_offsets(topic_to_partition_offset_map)

    def commit_offsets(self, topic_to_partition_offset_map):
        """Commits offset information to kafka.  Allows lower-level control for
        committing offsets.  In general, :meth:`commit_message` or
        :meth:`commit_messages` should be used, but this can be useful when paired with
        :meth:`data_pipeline.position_data.PositionData.topic_to_last_position_info_map`.

        **Example**::
            The `topic_to_partition_offset_map` should be formatted like::

                {
                  'topic1': {0: 83854, 1: 8943892},
                  'topic2': {0: 190898}
                }

        Args::
            topic_to_partition_offset_map (Dict[str, Dict[int, int]]): Maps from
                topics to a partition and offset map for each topic.
        """
        topic_to_partition_offset_map = self._get_offsets_map_to_be_committed(
            topic_to_partition_offset_map
        )
        return self._send_offset_commit_requests(
            offset_commit_request_list=[
                OffsetCommitRequest(
                    topic=kafka_bytestring(topic),
                    partition=partition,
                    offset=offset,
                    metadata=None
                ) for topic, partition_map in topic_to_partition_offset_map.iteritems()
                for partition, offset in partition_map.iteritems()
            ]
        )

    def _get_offsets_map_to_be_committed(
        self,
        topic_to_partition_offset_map
    ):
        filtered_topic_to_partition_offset_map = defaultdict(dict)
        for topic, partition_map in topic_to_partition_offset_map.items():
            for partition, offset in partition_map.items():
                if (partition not in self.topic_to_partition_offset_map_cache[topic] or
                        self.topic_to_partition_offset_map_cache[topic][partition] != offset):
                    self.topic_to_partition_offset_map_cache[topic][partition] = offset
                    filtered_topic_to_partition_offset_map[topic][partition] = offset
        return filtered_topic_to_partition_offset_map

    @contextmanager
    def ensure_committed(self, messages):
        """ Context manager which calls `commit_messages` on exit, useful for
        ensuring a group of messages are committed even if an error is
        encountered.

        **Example**::

            for msg in consumer:
                if msg is None:
                    continue
                with consumer.ensure_committed(msg):
                    do_things_with_message(msg)

        **Example**::

            with consumer.ensure_committed(consumer.get_messages()) as messages:
                do_things_with_messages(messages)

        Args:
            messages (list[data_pipeline.message.Message]): Messages to commit
                when this context manager exits. May also pass a single
                `data_pipeline.message.Message` object.
        """
        try:
            yield messages
        finally:
            if isinstance(messages, Message):
                messages = [messages]
            self.commit_messages(messages=messages)

    def reset_topics(self, topic_to_consumer_topic_state_map):
        """ Stop and restart the Consumer with a new
        topic_to_consumer_topic_state_map. This can be used to change topics
        which are being consumed and/or modifying the offsets of the topics
        already being consumed.

        **Example**::

            with Consumer(
                consumer_name='example',
                team_name='bam',
                expected_frequency_seconds=12345,
                topic_to_consumer_topic_state_map={'topic1': None}
            ) as consumer:
                while True:
                    messages = consumer.get_messages(
                        count=batch_size,
                        blocking=True,
                        timeout=batch_timeout
                    )
                    process_messages(messages)
                    if no_new_topics():
                        continue
                    consumer.commit_messages(messages)
                    topic_map = {}
                    for topic in consumer.topic_to_partition_map.keys():
                        topic_map[topic] = None
                    new_topics = get_new_topics()
                    for topic in new_topics:
                        if topic not in topic_map:
                            topic_map[topic] = None
                    consumer.reset_topics(topic_map)

        Note:
            This is an expensive operation, roughly equivalent to destroying
            and recreating the Consumer, so make sure you only are calling this
            when absolutely necessary.

            It's also important to note that you should probably be calling
            `commit_messages` just prior to calling this, otherwise the offsets
            may be lost for the topics you were previously tailing.

        Args:
            topic_to_consumer_topic_state_map ({str:Optional(ConsumerTopicState)}):
                A map of topic names to `ConsumerTopicState` objects which
                define the offsets to start from. The ConsumerTopicState of a topic
                may be `None`, in which case the committed kafka offset for the
                consumer_name is used. If there is no committed kafka offset
                for the consumer_name the Consumer will begin from the
                `auto_offset_reset` offset in the topic.
        """
        self.stop()
        self._commit_topic_offsets(topic_to_consumer_topic_state_map)
        self._set_topic_to_partition_map(topic_to_consumer_topic_state_map)
        self._start_consumer()

    def _commit_topic_offsets(self, topic_to_consumer_topic_state_map):
        if topic_to_consumer_topic_state_map:
            logger.info("Committing offsets for Consumer '{0}'...".format(
                self.client_name
            ))
            topic_to_partition_offset_map = {}
            for topic, consumer_topic_state in topic_to_consumer_topic_state_map.iteritems():
                if consumer_topic_state is None:
                    continue
                topic_to_partition_offset_map[topic] = consumer_topic_state.partition_offset_map
            self.commit_offsets(topic_to_partition_offset_map)
            logger.info("Offsets committed for Consumer '{0}'...".format(
                self.client_name
            ))

    def _send_offset_commit_requests(self, offset_commit_request_list):
        if len(offset_commit_request_list) > 0:
            retry_on_exception(
                self._consumer_retry_policy,
                (FailedPayloadsError),
                self.kafka_client.send_offset_commit_request,
                group=kafka_bytestring(self.client_name),
                payloads=offset_commit_request_list
            )

    @property
    def _region_cluster_config(self):
        """ The ClusterConfig for Kafka cluster to connect to. If cluster_name
        is not specified, it will default to the value set in Config"""
        if self.cluster_name:
            return discovery.get_kafka_cluster(
                cluster_type=self.cluster_type,
                client_id=self.client_name,
                cluster_name=self.cluster_name
            )
        else:
            return get_config().cluster_config

    @property
    def _kafka_consumer_config(self):
        """ The `KafkaConsumerConfig` for the Consumer.

        Notes:
            This is not a `@cached_property` since there is the possibility
            that the cluster_config could change during runtime and users could
            leverage this for responding to topology changes.

            `auto_commit` is set to False to ensure clients can determine when
            they want their topic offsets committed via commit_messages(..)
        """
        return KafkaConsumerConfig(
            group_id=self.client_name,
            cluster=self._region_cluster_config,
            auto_offset_reset=self.auto_offset_reset,
            auto_commit=False,
            partitioner_cooldown=self.partitioner_cooldown,
            use_group_sha=self.use_group_sha,
            pre_rebalance_callback=self.pre_rebalance_callback,
            post_rebalance_callback=self._apply_post_rebalance_callback_to_partition
        )

    def _apply_post_rebalance_callback_to_partition(self, partitions):
        """
        Removes the topics not present in the partitions list
        from the consumer topic_to_partition_map

        Args:
            partitions: List of current partitions the kafka
            broker holds
        """
        self.topic_to_partition_map = dict(partitions)
        self.reset_topic_to_partition_offset_cache()

        if self.post_rebalance_callback:
            return self.post_rebalance_callback(partitions)

    def refresh_new_topics(
        self,
        topic_filter=None,
        pre_topic_refresh_callback=None
    ):
        """
        Get newly created topics that match given criteria and currently not
        being tailed by the consumer and refresh internal topics list for these
        new topics.

        **Example**::

            topic_filter = TopicFilter(namespace_name='test_namespace')
            with Consumer(
                consumer_name='example',
                team_name='bam',
                expected_frequency_seconds=12345,
                topic_to_consumer_topic_state_map={'topic1': None, 'topic2': None}
            ) as consumer:
                while True:
                    messages = consumer.get_messages(
                        count=batch_size,
                        blocking=True,
                        timeout=batch_timeout
                    )
                    process_messages(messages)
                    if no_new_topics():
                        continue
                    consumer.commit_messages(messages)
                    consumer.refresh_new_topics(
                        topic_filter,
                        pre_topic_refresh_callback
                    )

        Note:
            pre_topic_refresh_callback can be used to perform custom logic and
            takes the currently tailed topics and the newly added topics as
            arguments. For example, in the aforementioned example, if a new
            topic 'topic3' is being added to namespace 'test_namespace' then
            `pre_topic_refresh_callback(['topic1', 'topic2'], ['topic3'])`
            would be executed.

            Commit the offsets of already consumed messages before refreshing
            to prevent duplicate messages. Else, the consumer will pick up the
            last committed offsets of topics and will consume already consumed
            messages.

        Args:
            topic_filter (Optional[TopicFilter]): criteria to filter newly
                created topics.
            pre_topic_refresh_callback:
                (Optional[Callable[[list[str], list[str]], Any]]): function
                that performs custom logic before the consumer start tailing
                new topics. The function will take a list of old_topic_names
                and list of new_topic_names. The return value of the function
                is ignored.

        Returns:
            [data_pipeline.schematizer_clientlib.models.Topic]: A list of new topics.
        """
        # TODO [DATAPIPE-843|clin] deprecate this function in favor of new
        # `refresh_topics` function once AST is revised to use the new function.
        new_topics = self._get_new_topics(topic_filter)
        new_topics = [topic for topic in new_topics
                      if topic.name not in self.topic_to_partition_map]

        if new_topics:
            new_topic_names = [topic.name for topic in new_topics]
            old_topic_names = self.topic_to_partition_map.keys()
            if pre_topic_refresh_callback:
                pre_topic_refresh_callback(old_topic_names, new_topic_names)
            self.stop()
            self._update_topic_to_partition_map(new_topic_names)
            self._start_consumer()

        return new_topics

    def _get_new_topics(self, topic_filter):
        new_topics = self._schematizer.get_topics_by_criteria(
            namespace_name=topic_filter.namespace_name,
            source_name=topic_filter.source_name,
            created_after=topic_filter.created_after
        )
        if topic_filter.filter_func:
            new_topics = topic_filter.filter_func(new_topics)
        return new_topics

    def _update_topic_to_partition_map(self, new_topics):
        for new_topic in new_topics:
            self.topic_to_partition_map[new_topic] = None

    def refresh_topics(self, consumer_source):
        """Refresh the topics this consumer is consuming from based on the
        settings in the given consumer_source.

        Args:
            consumer_source (data_pipeline.consumer_source.ConsumerSource): one
            of ConsumerSource types, such as SingleTopic, TopicsInFixedNamespaces.

        Returns:
            List[str]: A list of new topic names that this consumer starts to
            consume.
        """
        topics = consumer_source.get_topics()
        new_topics = [topic for topic in topics
                      if topic not in self.topic_to_partition_map]

        if new_topics:
            self.stop()
            self._update_topic_to_partition_map(new_topics)
            self._start_consumer()

            # If a new topic doesn't exist, when the consumer restarts, it will
            # be removed from the topic state map after the re-balance callback.
            # In such case, the non-existent topics are removed from the new_topics.
            new_topics = [topic for topic in new_topics
                          if topic in self.topic_to_partition_map]

        return new_topics


class TopicFilter(object):
    """Criteria to filter topics.

    Args:
        namespace_name (Optional[str]): filter topics by their namespace name.
        source_name (Optional[str]): filter topics by their source name. Note
            that same source name may appear in multiple namespaces.
        created_after (Optional[int]): get topics created after this timestamp.
            The topics created at the same timestamp are included as well.
        filter_func (Optional[function]): function that performs custom logic
            to filter topics.  The input of this function will be a list of
            topics already filtered by specified namespace, source, and/or
            created_after timestamp.  This function should return a list of
            filtered topics.
    """

    def __init__(
        self,
        namespace_name=None,
        source_name=None,
        created_after=None,
        filter_func=None
    ):
        self.namespace_name = namespace_name
        self.source_name = source_name
        self.created_after = created_after
        self.filter_func = filter_func



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import getpass
import os
import re
import subprocess
import time

import requests
from docker import Client
from kafka import KafkaClient
from kafka.common import KafkaUnavailableError

from data_pipeline.config import configure_from_dict
from data_pipeline.config import get_config


logger = get_config().logger


class ContainerUnavailableError(Exception):
    def __init__(self, project='unknown', service='unknown'):
        msg = open("logs/docker-compose.log", 'r').read()
        Exception.__init__(self, "Container for project {0} and service {1} failed to start with msg {2}".format(
            project,
            service,
            msg
        ))


class Containers(object):
    """Context manager that defers to already running docker
    containers if available, and if not, runs new containers for the duration
    of tests.

    Examples:

        To use the data pipeline containers in another project, include a
        fixture in `conftest.py` that will launch kafka, schematizer and
        dependent containers, then configure the clientlib to use them::

            @pytest.yield_fixture(scope='session')
            def containers():
                with Containers() as containers:
                    yield containers

        You'll also need to install the data_pipeline testing_helpers extras,
        to get all the necessary dependencies.  The easiest way to do that is
        to include the following in your packages `requirements-dev.txt`::

            data_pipeline[testing_helpers]

        :meth:`create_kafka_topic` can be used to create a new kafka topic in
        the running container::

            @pytest.fixture
            def topic(containers):
                containers.create_kafka_topic(str('some-topic-name'))

        When set to `true`, the `LEAVE_CONTAINERS_RUNNING` environment variable
        will prevent this class from shutting down testing containers between
        runs::

            $ LEAVE_CONTAINERS_RUNNING=true py.test tests/

        If you leave containers running, you'll need to manually clean up.  If
        you add a make target as described below, that can be accomplished
        by running something like::

            $ $(make compose-prefix) kill && $(make compose-prefix) rm

        The `FORCE_FRESH_CONTAINERS` environment variable will force this class
        to stop any already-running containers, remove them, and start them
        again from scratch::

            $ FORCE_FRESH_CONTAINERS=true py.test tests/

        The `PULL_CONTAINERS` environment variable updates all of the underlying
        containers before starting containers, if it's necessary to start new
        containers::

            $ PULL_CONTAINERS=true py.test tests/

        The `OPEN_SOURCE_MODE` environment variable should be set when running in
        open source mode::

            $ OPEN_SOURCE_MODE=true py.test tests/

        When running tests using an automated system like Jenkins, both
        `FORCE_FRESH_CONTAINERS` and `PULL_CONTAINERS` should be set, so tests
        are always run against pristene containers and the most recently deployed
        version of the schematizer. A Makefile target can set these variables
        directly::

            test:
                # This will timeout after 15 minutes, in case there is a hang on jenkins
                PULL_CONTAINERS=true FORCE_FRESH_CONTAINERS=true timeout -9 900 tox $(REBUILD_FLAG)

        An "Archive Artifacts" step can be added to the automated build
        capturing the projects "logs/**/*", which will contain the projects
        docker-compose log.  Additional debugging logs for the test run can be
        captured by adding the following to `conftest.py`::

            logging.basicConfig(
                level=logging.DEBUG,
                filename='logs/test.log',
                format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s'
            )

        If you want to manually inspect the containers with docker-compose, you
        can get the preconfigured docker-compose command by running something
        like the following from within a virtualenv::

            python -c "from data_pipeline.testing_helpers.containers import Containers; print Containers.compose_prefix()"

        That will return something along the lines of::

            docker-compose --file=/nail/home/justinc/pg/data_pipeline/data_pipeline/testing_helpers/docker-compose.yml --project-name=datapipelinejustinc  # NOQA

        In this project, a make target wraps this, which allows
        users to run commands like::

            $(make compose-prefix) ps
    """

    @classmethod
    def compose_prefix(cls):
        """Returns a configured docker-compose command for the user and project
        """
        return "docker-compose {}".format(Containers()._compose_options)

    @classmethod
    def get_container_info(cls, project, service, docker_client=None):
        """Returns container information for the first container matching project
        and service.
        Raises ContainerUnavailableError if the container is unavailable.

        Args:
            project: Name of the project the container is hosting
            service: Name of the service that the container is hosting
            docker_client: The docker client object
        """
        if docker_client is None:
            docker_client = Client(version='auto')

        try:
            return next(
                c for c in docker_client.containers() if
                c['Labels'].get('com.docker.compose.project') == project and
                c['Labels'].get('com.docker.compose.service') == service
            )
        except StopIteration:
            raise ContainerUnavailableError(project=project, service=service)

    @classmethod
    def get_container_ip_address(cls, project, service, timeout_seconds=100):
        """Fetches the ip address assigned to the running container.
        Throws ContainerUnavailableError if the retry limit is reached.

        Args:
            project: Name of the project the container is hosting
            service: Name of the service that the container is hosting
            timeout_seconds: Retry time limit to wait for containers to start
                             Default in seconds: 60
        """
        docker_client = Client(version='auto')
        container_id = cls._get_container_id(project, service, docker_client, timeout_seconds)
        return docker_client.inspect_container(
            container_id
        )['NetworkSettings']['IPAddress']

    @classmethod
    def exec_command(cls, command, project, service, timeout_seconds=60):
        """Executes the command in the project and service container running under
        docker-compose.
        Throws ContainerUnavailableError if the retry limit is reached.

        Args:
            command: The command to be executed in the container
            project: Name of the project the container is hosting
            service: Name of the service that the container is hosting
            timeout_seconds: Retry time limit to wait for containers to start
                             Default in seconds: 60
        """
        docker_client = Client(version='auto')
        container_id = cls._get_container_id(project, service, docker_client, timeout_seconds)
        exec_id = docker_client.exec_create(container_id, command)['Id']
        docker_client.exec_start(exec_id)

    @classmethod
    def get_kafka_connection(cls, timeout_seconds=15):
        """Returns a kafka connection, waiting timeout_seconds for the container
        to come up.

        Args:
            timeout_seconds: Retry time (seconds) to get a kafka connection
        """
        end_time = time.time() + timeout_seconds
        logger.info("Getting connection to Kafka container on yocalhost")
        while end_time > time.time():
            try:
                return KafkaClient(get_config().cluster_config.broker_list)
            except KafkaUnavailableError:
                logger.info("Kafka not yet available, waiting...")
                time.sleep(0.1)
        raise KafkaUnavailableError()

    @classmethod
    def _get_container_id(cls, project, service, docker_client, timeout_seconds=60):
        end_time = time.time() + timeout_seconds
        while end_time > time.time():
            try:
                container_info = cls.get_container_info(
                    project,
                    service,
                    docker_client=docker_client
                )
                return container_info['Id']
            except ContainerUnavailableError:
                time.sleep(1)
        else:
            raise ContainerUnavailableError(project=project, service=service)

    def __init__(self, additional_compose_file=None, additional_services=None):
        # To resolve docker client server version mismatch issue.
        os.environ["COMPOSE_API_VERSION"] = "auto"
        dir_name = os.path.split(os.getcwd())[-1]
        self.project = "{}{}".format(
            re.sub(r'[^a-z0-9]', '', dir_name.lower()),
            getpass.getuser()
        )
        self.additional_compose_file = additional_compose_file

        self.services = ["zookeeper", "schematizer", "kafka"]

        if additional_services is not None:
            self.services.extend(additional_services)

        # This variable is meant to capture the running/not-running state of
        # the dependent testing containers when tests start running.  The idea
        # is, we'll only start and stop containers if they aren't already
        # running.  If they are running, we'll just use the ones that exist.
        # It takes a while to start all the containers, so when running lots of
        # tests, it's best to start them out-of-band and leave them up for the
        # duration of the session.
        self.containers_already_running = self._are_containers_already_running()

    def __enter__(self):
        if (
            not self.containers_already_running or
            self._is_envvar_set('FORCE_FRESH_CONTAINERS')
        ):
            self._start_containers()
        else:
            logger.info("Using running containers")

        self.use_testing_containers()

        return self

    def __exit__(self, type, value, traceback):
        if not (
            self.containers_already_running or
            self._is_envvar_set('LEAVE_CONTAINERS_RUNNING')
        ):
            # only stop containers that we started
            self._stop_containers()
        return False  # Don't Suppress Exception

    def create_kafka_topic(self, topic):
        """This method execs in the docker container because it's the only way to
        control how the topic is created.

        Args:
            topic (str): Topic name to create
        """
        conn = Containers.get_kafka_connection()
        if conn.has_metadata_for_topic(topic):
            return

        logger.info("Creating Fake Topic")
        if not isinstance(topic, str):
            raise ValueError("topic must be a str, it cannot be unicode")

        kafka_create_topic_command = (
            "$KAFKA_HOME/bin/kafka-topics.sh --create --zookeeper zk:2181 "
            "--replication-factor 1 --partition 1 --topic {topic}"
        ).format(topic=topic)

        Containers.exec_command(kafka_create_topic_command, self.project, 'kafka')

        logger.info("Waiting for topic")
        conn.ensure_topic_exists(
            topic,
            timeout=get_config().topic_creation_wait_timeout
        )
        conn.close()
        logger.info("Topic Exists")
        assert conn.has_metadata_for_topic(topic)

    @property
    def _compose_options(self):
        file_name = ("docker-compose-opensource.yml"
                     if self._is_envvar_set('OPEN_SOURCE_MODE')
                     else "docker-compose.yml")
        compose_file = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                file_name
            )
        )
        if self.additional_compose_file:
            # The additional compose file need to be put in front.
            return "--file={} --file={} --project-name={}".format(
                self.additional_compose_file, compose_file, self.project
            )
        else:
            return "--file={} --project-name={}".format(compose_file, self.project)

    def use_testing_containers(self):
        """Configures the data pipeline clientlib to use the containers"""
        schematizer_ip = Containers.get_container_ip_address(self.project, 'schematizer')
        schematizer_host_and_port = "{}:8888".format(schematizer_ip)

        zookeeper_ip = Containers.get_container_ip_address(self.project, 'zookeeper')
        zookeeper_host_and_port = "{}:2181".format(zookeeper_ip)

        kafka_ip = Containers.get_container_ip_address(self.project, 'kafka')
        kafka_host_and_port = "{}:9092".format(kafka_ip)

        configure_from_dict(dict(
            schematizer_host_and_port=schematizer_host_and_port,
            kafka_zookeeper=zookeeper_host_and_port,
            kafka_broker_list=[kafka_host_and_port],
            should_use_testing_containers=True
        ))

    def _are_containers_already_running(self):
        return all(self._is_service_running(service) for service in self.services)

    def _is_service_running(self, service):
        return int(subprocess.call(
            "docker-compose {} ps {} | grep Up".format(self._compose_options, service),
            shell=True,
            stdout=subprocess.PIPE
        )) == 0

    def _start_containers(self):
        self._stop_containers()
        if self._is_envvar_set('PULL_CONTAINERS'):
            logger.info("Updating Containers")
            self._run_compose('pull')
        if self.additional_compose_file:
            logger.info("Building Containers")
            self._run_compose('build')
        logger.info("Starting Containers")
        self._run_compose('up', '-d', *self.services)
        self._wait_for_services()

    def _wait_for_services(self, timeout_seconds=60):
        self.use_testing_containers()
        self._wait_for_schematizer(timeout_seconds=90)
        self._wait_for_kafka(timeout_seconds)

    def _wait_for_schematizer(self, timeout_seconds):
        # wait for schematizer to pass health check
        end_time = time.time() + timeout_seconds
        logger.info("Waiting for schematizer to pass health check")
        count = 0
        while end_time > time.time():
            time.sleep(0.1)
            try:
                r = requests.get(
                    "http://{0}/v1/namespaces".format(get_config().schematizer_host_and_port)
                )
                if 200 <= r.status_code < 300:
                    count += 1
                    if count >= 2:
                        return
            except Exception:
                count = 0
            finally:
                logger.info("Schematizer not yet available, waiting...")
        raise ContainerUnavailableError(project='schematizer', service='schematizer')

    def _wait_for_kafka(self, timeout_seconds):
        # This will raise an exception after timeout_seconds, if it can't get
        # a connection
        Containers.get_kafka_connection(timeout_seconds)

    def _stop_containers(self):
        logger.info("Stopping containers")
        self._run_compose('kill')
        self._run_compose('rm', '--force')

    def _run_compose(self, *args):
        args = list(args)

        if not os.path.isdir('logs'):
            os.mkdir('logs')

        with open("logs/docker-compose.log", "a") as f:
            subprocess.call(
                ['docker-compose'] + self._compose_options.split() + args,
                stdout=f,
                stderr=subprocess.STDOUT
            )

    def _is_envvar_set(self, envvar):
        return os.getenv(envvar, 'true').lower() in ['t', 'true', 'y', 'yes']



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
This script will create a JSON version of the Google Spreadsheet schema reference.

Google Spreadsheet:
https://docs.google.com/spreadsheets/d/1ZIE8UdMadTgBpcELOnGLNJIhOLXBsij0_f5BiwlAQss/edit#gid=8

Usage:
    1. download sheets as csvs from the Google Spreadsheet into the same directory as this script
        - ref_cols.csv          Sheet 1 ("DW Schema Reference")
        - ref_tables.csv        Sheet 2 ("DW Tables")
    2. download BAM's listings of table owners into the same directory as this script
        - ref_owners.csv        Original Version
        - ref_owners_new.csv    Updated Version
    3. run the script and the output will be in schema_ref.json
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import csv
import json

from data_pipeline.helpers.frozendict_json_encoder import FrozenDictEncoder


def _read_rows_from_file(file_name):

    rows = []
    with open(file_name, 'rb') as file:
        reader = csv.reader(file)
        for row in reader:
            rows.append(row)
        return rows


def _parse_col_row(row):

    _, _, col_name, pos, _, nullable, write_once, data_type, _, _, _, _, description, _, notes = row

    if nullable == 'NO':
        data_type += ' not null'
    if write_once == 'YES':
        data_type += ' write once'

    if notes.strip() == '0':
        notes = ''

    return {
        'name': col_name,
        'doc': description,
        'note': notes,
    }


if __name__ == '__main__':

    owners_rows = _read_rows_from_file('ref_owners.csv')
    owners_new_rows = _read_rows_from_file('ref_owners_new.csv')
    tables_rows = _read_rows_from_file('ref_tables.csv')
    cols_rows = _read_rows_from_file('ref_cols.csv')

    tables_rows = tables_rows[1:]

    output = {
        'doc_source': 'https://docs.google.com/spreadsheets/d/1ZIE8UdMadTgBpcELOnGLNJIhOLXBsij0_f5BiwlAQss/edit#gid=11',
        'doc_owner': 'bam@yelp.com',
        'docs': []
    }

    for row in tables_rows:

        schema, name, category, description, _, _, notes, _ = row
        table_output = {
            'namespace': schema,
            'source': name,
            'doc': description,
            'note': notes,
            'category': category,
            'fields': []
        }

        owner_row = filter(lambda row: row[3] == name, owners_rows)
        owner_new_row = filter(lambda row: row[3] == name, owners_new_rows)

        try:
            _, source_path, _, _, _, owner = owner_row.pop()
            owner = owner.split(',')[0]

            table_output['owner_email'] = owner
            table_output['file_display'] = source_path
            table_output['file_url'] = 'https://opengrok.yelpcorp.com/xref/yelp-main/' + source_path

        except IndexError:

            if len(owner_new_row) > 0:

                _, source_path, _, _, _, owner = owner_new_row.pop()
                owner = owner.split(',')[0]

                table_output['owner_email'] = owner
                table_output['file_display'] = source_path
                table_output['file_url'] = 'https://opengrok.yelpcorp.com/xref/yelp-main/' + source_path

            else:

                table_output['owner_email'] = ''
                table_output['file_display'] = ''
                table_output['file_url'] = ''

        col_rows = filter(lambda row: row[0] == schema and row[1] == name, cols_rows)

        for row in col_rows:
            table_output['fields'].append(_parse_col_row(row))

        output['docs'].append(table_output)

    with open('schema_ref.json', 'wb') as outfile:
        outfile.write(json.dumps(output, cls=FrozenDictEncoder))
        outfile.close()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import copy
import logging
from datetime import timedelta

import pysensu_yelp

from data_pipeline.tools.heartbeat_periodic_processor import BasePeriodicProcessor


class SensuAlertManager(BasePeriodicProcessor):
    """ This class triggers sensu alert if the producer falls behind real time.
    The check will be triggered every interval_in_seconds seconds to avoid flapping and
    overflowing sensu servers. If we fall behind and then recover, sensu will take care
    of resolving this alert itself.  Typically the producer will call process on this
    in this class for every message produced.

    Note the check_every variable in the sensu alert is defaulted to interval_in_seconds
    if it's not specified.

    Args:
        interval_in_seconds(int): check_every interval for sensu
        result_dict(dict): dictionary with sensu parameters.  For details see
             http://pysensu-yelp.readthedocs.io/en/latest/index.html?highlight=send_event for details
        service_name(str): name of the service possibly with underscores
        disable(bool): disable this alert manager
        max_delay_seconds(int): number of seconds before timing out
    """

    def __init__(
        self,
        interval_in_seconds,
        service_name,
        result_dict,
        max_delay_seconds,
        disable=False
    ):
        super(SensuAlertManager, self).__init__(interval_in_seconds)
        self._service_name = service_name
        self._setup_ok_result_dict(result_dict)
        self._setup_delayed_result_dict()
        self._setup_disabled_alert_dict()
        self._log = logging.getLogger('{}.util.sensu_alert_manager'.format(service_name))
        self._disable = disable
        self._should_send_sensu_disabled_message = False
        self._max_delay = timedelta(seconds=max_delay_seconds)

    def _setup_ok_result_dict(self, result_dict):
        self._ok_result_dict = result_dict
        self._ok_result_dict['check_every'] = result_dict.get(
            'check_every',
            '{time}s'.format(time=self.interval_in_seconds)
        )
        self._ok_result_dict.update({
            'status': 0,
            'output': '{} has caught up to real time'.format(
                self._service_name
            )
        })

    def _setup_delayed_result_dict(self):
        self._delayed_result_dict = copy.deepcopy(self._ok_result_dict)
        self._delayed_result_dict.update({'status': 2})

    def _setup_disabled_alert_dict(self):
        self._disabled_alert_dict = copy.deepcopy(self._ok_result_dict)
        self._disabled_alert_dict.pop("ttl")
        self._disabled_alert_dict.update({
            'output': 'disabling sensu alert for {}'.format(self._service_name)
        })

    @property
    def disable(self):
        return self._disable

    @disable.setter
    def disable(self, new_value):
        # This condition means that something has toggled the disable variable from
        # False to True.  In this case we need to send sensu a message without a ttl
        # so it stops looking for ttl's.  If we didn't do this sensu would not get
        # any OK's so would keep alerting even if self._disable is True
        if not self._disable and new_value:
            self.log_and_send_event(self._disabled_alert_dict)
        self._disable = new_value

    def process(self, timestamp=None):
        if timestamp is None or self.disable:
            return

        # This timestamp param has to be timezone aware, otherwise it will not be
        # able to compare with timezone aware timestamps.
        delay_time = self._utc_now - timestamp
        if delay_time <= self._max_delay:
            self.log_and_send_event(self._ok_result_dict)
            return

        self._delayed_result_dict.update({
            'output': '{service} is falling {delay_time} min behind real time'.format(
                service=self._service_name,
                delay_time=delay_time),
        })
        self.log_and_send_event(self._delayed_result_dict)

    def log_and_send_event(self, output_dict):
        self._log.info("{} status: {}, output: {}.".format(
            self._service_name,
            output_dict['status'],
            output_dict['output'])
        )
        pysensu_yelp.send_event(**output_dict)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from data_pipeline.tools.introspector.base_command import IntrospectorCommand


class _BaseListCommand(IntrospectorCommand):

    @classmethod
    def add_base_arguments(cls, parser):
        super(_BaseListCommand, cls).add_base_arguments(parser)

        parser.add_argument(
            "-s",
            "--sort-by",
            type=str,
            default=None,
            help="Sort the listing by a particular field of the object "
                 "in ascending order (by default)"
        )

        parser.add_argument(
            "--descending-order", "--desc",
            action="store_true",
            default=False,
            help="Use --sort-by with descending order (Will be ignored if --sort-by is not set)"
        )

    @classmethod
    def get_description(cls):
        return "List {}, as a JSON array of formatted {}. Fields: {}".format(
            cls.list_type, cls.list_type, cls.fields
        )

    def process_args(self, args, parser):
        super(_BaseListCommand, self).process_args(args, parser)
        self.sort_by = args.sort_by
        self.descending_order = args.descending_order
        if self.sort_by and self.sort_by not in self.fields:
            raise parser.error(
                "You can not sort_by by {} for list type {}. Possible fields are: {}".format(
                    self.sort_by, self.list_type, self.fields
                )
            )



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from uuid import UUID

from kafka_utils.util import offsets

from data_pipeline._fast_uuid import FastUUID
from data_pipeline.base_consumer import ConsumerTopicState
from data_pipeline.config import get_config
from data_pipeline.consumer import Consumer
from data_pipeline.expected_frequency import ExpectedFrequency

logger = get_config().logger


def get_first_offset_at_or_after_start_timestamp(kafka_client, topics, start_timestamp):
    """Uses binary search to find the first offset that comes after start_timestamp for each
    topic in topics. If multiple items are present for a timestamp, the first one (closer to
    low_mark) is returned back.

    Outputs a result_topic_to_consumer_topic_state_map which can be used to set offsets

    :param kafka_client: kafka client to be used for getting watermarks and binary search.
    :param topics: a list of topics. eg. ['test_topic_1', 'test_topic_2']
    :param start_timestamp: epoch timestamp eg. 1463086536

    :returns: a dict mapping topic to the nearest starting timestamp.
              eg.
              {'test_topic_1': ConsumerTopicState({0: 43}, None),
              'test_topic_2': ConsumerTopicState({0: 55, 1: 32}, None)}
    """
    watermarks = offsets.get_topics_watermarks(
        kafka_client,
        topics,
        raise_on_error=False
    )

    topic_to_consumer_topic_state_map = _build_topic_to_consumer_topic_state_map(watermarks)
    topic_to_range_map = _build_topic_to_range_map(watermarks)
    result_topic_to_consumer_topic_state_map = _build_empty_topic_to_consumer_topic_state_map(topics)

    _move_finished_topics_to_result_map(
        topic_to_consumer_topic_state_map,
        topic_to_range_map,
        result_topic_to_consumer_topic_state_map
    )
    while topic_to_consumer_topic_state_map:
        _get_message_and_alter_range(
            start_timestamp,
            topic_to_consumer_topic_state_map,
            topic_to_range_map,
            result_topic_to_consumer_topic_state_map
        )

    logger.info(
        "Got topic offsets based on start-date: {}".format(result_topic_to_consumer_topic_state_map)
    )
    return result_topic_to_consumer_topic_state_map


def _build_topic_to_range_map(watermarks):
    """Builds a topic_to_range_map from a kafka get_topics_watermarks response"""
    return {
        topic: {
            partition: {
                'high': marks.highmark,
                'low': marks.lowmark
            }
            for partition, marks in watermarks_map.items()
        }
        for topic, watermarks_map in watermarks.items()
    }


def _build_topic_to_consumer_topic_state_map(watermarks):
    """Builds a topic_to_consumer_topic_state_map from a kafka get_topics_watermarks response"""
    return {
        topic: ConsumerTopicState({
            partition: int((marks.highmark + marks.lowmark) / 2)
            for partition, marks in watermarks_map.items()
        }, None)
        for topic, watermarks_map in watermarks.items()
    }


def _build_empty_topic_to_consumer_topic_state_map(topics):
    """Builds a topic_to_consumer_topic_state_map from a list of topics where all of the interior
    partition_offset_maps are empty"""
    return {
        topic: ConsumerTopicState({}, None)
        for topic in topics
    }


def _get_message_and_alter_range(
    start_timestamp,
    topic_to_consumer_topic_state_map,
    topic_to_range_map,
    result_topic_to_consumer_topic_state_map
):
    """Create a consumer based on our topic_to_consumer_state_map, get a message, and based
    on that message's timestamp, adjust our topic ranges and maps with _update_ranges_for_message and
    _move_finisehd_topics_to_result_map"""
    # We create a new consumer each time since it would otherwise require refactoring how
    # we consume from KafkaConsumerGroups (currently use an iterator, which doesn't support big jumps in offset)
    with Consumer(
        'data_pipeline_tailer_starting_offset_getter-{}'.format(
            str(UUID(bytes=FastUUID().uuid4()).hex)
        ),
        'bam',
        ExpectedFrequency.constantly,
        topic_to_consumer_topic_state_map
    ) as consumer:
        message = consumer.get_message(timeout=0.1, blocking=True)
        if message is None:
            return
        _update_ranges_for_message(
            message,
            start_timestamp,
            topic_to_consumer_topic_state_map,
            topic_to_range_map
        )
    _move_finished_topics_to_result_map(
        topic_to_consumer_topic_state_map,
        topic_to_range_map,
        result_topic_to_consumer_topic_state_map
    )


def _move_finished_topics_to_result_map(
    topic_to_consumer_topic_state_map,
    topic_to_range_map,
    result_topic_to_consumer_topic_state_map
):
    """Moves all "finished" (where max of topic range is equal to min) partitions
    to result_topic_to_consumer_topic_state_map for output, and removes them from the
    intermediate ones to stop getting messages from that partition"""
    for topic, consumer_topic_state_map in topic_to_consumer_topic_state_map.iteritems():
        for partition in consumer_topic_state_map.partition_offset_map.keys():
            topic_range = topic_to_range_map[topic][partition]
            if topic_range['high'] == topic_range['low']:
                result_topic_to_consumer_topic_state_map[topic].partition_offset_map[partition] = \
                    consumer_topic_state_map.partition_offset_map.pop(partition)

    for topic in topic_to_consumer_topic_state_map.keys():
        consumer_topic_state_map = topic_to_consumer_topic_state_map[topic]
        if not consumer_topic_state_map.partition_offset_map:
            topic_to_consumer_topic_state_map.pop(topic)


def _update_ranges_for_message(
    message,
    start_timestamp,
    topic_to_consumer_topic_state_map,
    topic_to_range_map
):
    """Given a message, updates our topic_to_range_maps and sets the new state of the
    topic_to_consumer_topic_state_map, based on start_timestamp"""
    topic = message.topic
    offset = message.kafka_position_info.offset
    partition = message.kafka_position_info.partition
    timestamp = message.timestamp

    if timestamp < start_timestamp:
        topic_to_range_map[topic][partition]['low'] = offset + 1
    else:
        topic_to_range_map[topic][partition]['high'] = offset
    new_mid = int(
        (
            topic_to_range_map[topic][partition]['high'] +
            topic_to_range_map[topic][partition]['low']
        ) / 2
    )
    topic_to_consumer_topic_state_map[topic].partition_offset_map[partition] = new_mid
    logger.debug(
        "During start-date offset search got message from "
        "topic|offset|part: {}|{}|{}, new range: {}".format(
            topic, offset, partition, topic_to_range_map[topic]
        )
    )



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import yaml

from data_pipeline.config import get_config


class Team(object):
    """Contains some helper methods for dealing with the data pipeline team
    configuration.  Eventually this class should encapsulate information about
    each team.  That's overkill for the current use-case, so it's not
    implemented yet.

    """

    @classmethod
    def config(cls):
        """Loads and decodes the
        :attr:`data_pipeline.config.Config.data_pipeline_teams_config_file_path`.

        TODO(justinc|DATAPIPE-348): Cache team config, dealing with invalidation
        when configuration changes.

        Returns:
            dict: team configuration
        """
        config_path = get_config().data_pipeline_teams_config_file_path
        return yaml.load(open(config_path).read())

    @classmethod
    def team_names(cls):
        """Lists all data pipeline teams

        Returns:
            list of str: all valid data pipeline team names
        """
        return cls.config()['teams'].keys()

    @classmethod
    def exists(cls, team_name):
        """Determines if a team exists, by name.

        Returns:
            bool: True if team_name exists for a valid team, false otherwise
        """
        return team_name in cls.team_names()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from data_pipeline.tools.introspector.list_command.namespaces import NamespacesListCommand
from data_pipeline.tools.introspector.list_command.sources import SourcesListCommand
from data_pipeline.tools.introspector.list_command.topics import TopicsListCommand


class ListCommandParser(object):

    @classmethod
    def add_parser(cls, subparsers):
        list_command_parser = subparsers.add_parser(
            "list",
            description="Get a list of specified items as a JSON array of objects."
        )

        list_command_subparsers = list_command_parser.add_subparsers()
        TopicsListCommand.add_parser(list_command_subparsers)
        SourcesListCommand.add_parser(list_command_subparsers)
        NamespacesListCommand.add_parser(list_command_subparsers)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals


# This metaclass is from http://stackoverflow.com/questions/6760685/creating-a-singleton-in-python
class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import simplejson

from data_pipeline.tools.introspector.base_command import IntrospectorCommand
from data_pipeline.tools.introspector.models import IntrospectorSchema


class _BaseRegisterCommand(IntrospectorCommand):

    @classmethod
    def add_base_arguments(cls, parser):
        super(_BaseRegisterCommand, cls).add_base_arguments(parser)
        cls.add_source_and_namespace_arguments(parser)

        parser.add_argument(
            "--source_owner_email",
            type=str,
            required=True,
            help="The email of the owner of the given source."
        )

        parser.add_argument(
            "--contains-pii",
            dest="pii",
            default=False,
            action="store_true",
            help="Flag indicating if schema contains pii. More info at y/pii"
        )

    def process_args(self, args, parser):
        super(_BaseRegisterCommand, self).process_args(args, parser)
        self.process_source_and_namespace_args(args, parser)
        self.source_owner_email = args.source_owner_email
        self.pii = args.pii

    def print_schema(self, schema):
        schema_dict = IntrospectorSchema(
            schema,
            include_topic_info=True
        ).to_ordered_dict()
        print simplejson.dumps(schema_dict)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import ctypes.util
import uuid

from cffi import FFI

from data_pipeline.config import get_config


class _UUIDBase(object):

    def uuid1(self):
        raise NotImplementedError()

    def uuid4(self):
        raise NotImplementedError()


class _LibUUID(_UUIDBase):
    """Fast c-wrapper for for uuid generation

    This class wraps the libuuid (http://linux.die.net/man/3/libuuid)
    C library to generate UUIDs much more quickly than with CPython alone.  The
    cffi library was chosen over ctypes because it's the preferred extension
    method under PyPy, and is consequently about twice as fast.

    python-libuuid won't compile with PyPy, which is why this class was
    created.  This code will work with both PyPy and CPython.

    **Benchmarks**:

    Bottom line - using this UUID4 implementation with PyPy is over 15 times
    faster than using CPython's UUID1 implementation.

    Using PyPy 2.5.0::

        FastUUID UUID1
        Total Time (100,000 iters): 3.87426400185 seconds
        Rate: 25,811.35/second

        Python UUID1
        Total Time (100,000 iters): 4.65500807762 seconds
        Rate: 21,482.24/second

        FastUUID UUID4
        Total Time (100,000 iters): 0.259171009064 seconds
        Rate: 385,845.63/second

        Python UUID4
        Total Time (100,000 iters): 0.626512765884 seconds
        Rate: 159,613.67/second

    Using CPython 2.6.7::

        FastUUID UUID1
        Total Time (100,000 iters): 0.798195838928 seconds
        Rate: 125,282.54/second

        Python UUID1
        Total Time (100,000 iters): 4.16052007675 seconds
        Rate: 24,035.46/second

        FastUUID UUID4
        Total Time (100,000 iters): 0.395098209381 seconds
        Rate: 253,101.63/second

        Python UUID4
        Total Time (100,000 iters): 3.39745283127 seconds
        Rate: 29,433.82/second
    """

    _ffi = None
    _libuuid = None

    def __init__(self):
        # Store these on the class since they should only ever be called once
        if _LibUUID._ffi is None or _LibUUID._libuuid is None:
            _LibUUID._ffi = FFI()

            # These definitions are from uuid.h
            _LibUUID._ffi.cdef("""
                typedef unsigned char uuid_t[16];

                void uuid_generate(uuid_t out);
                void uuid_generate_random(uuid_t out);
                void uuid_generate_time(uuid_t out);
            """)

            # By opening the library with dlopen, the compile step is skipped
            # dodging a class of errors, since headers aren't needed, just the
            # installed library.
            _LibUUID._libuuid = _LibUUID._ffi.dlopen(
                ctypes.util.find_library("uuid")
            )

            get_config().logger.debug(
                "FastUUID Created - FFI: ({}), LIBUUID: ({})".format(
                    _LibUUID._ffi,
                    _LibUUID._libuuid
                )
            )

        # Keeping only one copy of this around does result in
        # pretty substantial performance improvements - in the 10,000s of
        # messages per second range
        self.output = _LibUUID._ffi.new("uuid_t")

    def uuid1(self):
        _LibUUID._libuuid.uuid_generate_time(self.output)
        return self._get_output_bytes()

    def uuid4(self):
        _LibUUID._libuuid.uuid_generate_random(self.output)
        return self._get_output_bytes()

    def _get_output_bytes(self):
        return bytes(_LibUUID._ffi.buffer(self.output))


class _DefaultUUID(_UUIDBase):
    """Built-in uuid.
    """

    def uuid1(self):
        return uuid.uuid1().bytes

    def uuid4(self):
        return uuid.uuid4().bytes


class FastUUID(object):
    """It uses the fast c-wrapper (:class: data_pipeline._fast_uuid._FastUUID)
    for uuid generation, and falls back to use built-in uuid if that's not
    available.
    """

    _avail_uuids = [_LibUUID, _DefaultUUID]

    def __init__(self):
        for avail_uuid in self._avail_uuids:
            try:
                self._uuid_in_use = avail_uuid()
                break
            except Exception:
                get_config().logger.error(
                    "libuuid is unavailable, falling back to the slower built-in "
                    "uuid implementation.  On ubuntu, apt-get install uuid-dev."
                )

    def uuid1(self):
        """Generates a uuid1 - a device specific uuid

        Returns:
            bytes: 16-byte uuid
        """
        return self._uuid_in_use.uuid1()

    def uuid4(self):
        """Generates a uuid4 - a random uuid

        Returns:
            bytes: 16-byte uuid
        """
        return self._uuid_in_use.uuid4()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import datetime
import logging
from collections import OrderedDict
from contextlib import contextmanager

from bravado.exception import HTTPNotFound
from cached_property import cached_property
from kafka import KafkaClient
from kafka_utils.util import offsets
from kafka_utils.util.zookeeper import ZK

from data_pipeline.config import get_config
from data_pipeline.schematizer_clientlib.schematizer import get_schematizer
from data_pipeline.servlib.config_util import load_package_config
from data_pipeline.tools.introspector.models import IntrospectorNamespace
from data_pipeline.tools.introspector.models import IntrospectorSource
from data_pipeline.tools.introspector.models import IntrospectorTopic


class IntrospectorCommand(object):
    """The Data Pipeline Introspector provides information in to the current
    state of the data pipeline using the schematizer, zookeeper, and kafka.

    This is the base class of all of the individual introspection batches.
    """

    def __init__(self, log_name):
        self.log_name = log_name
        load_package_config('/nail/srv/configs/data_pipeline_tools.yaml')
        self.config = get_config()
        self.log = logging.getLogger(self.log_name)
        self._setup_logging()
        self.schematizer = get_schematizer()

    @classmethod
    def add_parser(cls, subparsers):
        raise NotImplementedError("All InspectorBatch sub classes must have add_parser defined")

    @classmethod
    def add_base_arguments(cls, parser):
        parser.add_argument(
            "-h",
            "--help",
            action="help",
            help="Show help messages and exit."
        )

        parser.add_argument(
            "-v", "--verbose",
            dest="verbosity",
            default=0,
            action="count",
            help="Adjust log level"
        )

    @classmethod
    def add_source_and_namespace_arguments(cls, parser):
        parser.add_argument(
            "--source-id",
            type=int,
            required=False,
            default=None,
            help="Source id of source to check against"
        )

        parser.add_argument(
            "--source-name",
            type=str,
            required=False,
            default=None,
            help="Source name of source to check against. "
                 "Must be used with --namespace"
        )

        parser.add_argument(
            "--namespace",
            required=False,
            type=str,
            default=None,
            help="Namespace name that contains a source of source name given. "
                 "If --source-id is given, then this will be ignored."
        )

    @cached_property
    def _kafka_topics(self):
        with ZK(self.config.cluster_config) as zk:
            return zk.get_topics(
                names_only=True,
                fetch_partition_state=False
            )

    @cached_property
    def _all_topic_watermarks(self):
        topics = self._kafka_topics
        with self._kafka_client() as kafka_client:
            return offsets.get_topics_watermarks(
                kafka_client,
                topics
            )

    @contextmanager
    def _kafka_client(self):
        kafka_client = KafkaClient(self.config.cluster_config.broker_list)
        try:
            yield kafka_client
        finally:
            kafka_client.close()

    def _retrieve_names_from_source_id(self, source_id):
        """Returns (source_name, namespace_name) of source with given source_id"""
        source = self.schematizer.get_source_by_id(source_id)
        return (source.name, source.namespace.name)

    def process_source_and_namespace_args(self, args, parser):
        self.source_id = args.source_id
        self.source_name = args.source_name
        self.namespace = args.namespace
        if self.source_id:
            self.source_name, self.namespace = self._retrieve_names_from_source_id(
                self.source_id
            )
            if args.namespace:
                self.log.warning(
                    "Since source id was given, --namespace will be ignored"
                )
        else:
            if not self.namespace:
                raise parser.error(
                    "--namespace must be provided when given a source name as source identifier."
                )

    def _does_topic_range_map_have_messages(self, range_map):
        range_sum = 0
        for _, watermark_range in range_map.iteritems():
            range_sum += watermark_range
        return range_sum > 0

    def _topic_watermarks_to_range(self, watermarks):
        return {
            partition: partition_offsets.highmark - partition_offsets.lowmark
            for partition, partition_offsets in watermarks.items()
        }

    def _topic_to_range_map_to_topics_list(self, topic_to_range_map):
        output = []
        for topic, range_map in topic_to_range_map.iteritems():
            try:
                topic_result = self.schematizer.get_topic_by_name(topic)._asdict()
                topic_result['range_map'] = range_map
                output.append(topic_result)
            except HTTPNotFound:
                pass
        return output

    @cached_property
    def _topics_with_messages_to_range_map(self):
        topic_to_range_map = {
            topic: self._topic_watermarks_to_range(topic_watermarks)
            for topic, topic_watermarks in self._all_topic_watermarks.items()
        }
        return {
            topic: range_map
            for topic, range_map in topic_to_range_map.items()
            if (self._does_topic_range_map_have_messages(range_map))
        }

    @cached_property
    def active_topics(self):
        """A cached list of active topics,
        meaning that they are in both kafka and the schematizer,
        and have at least one message.

        Each topic include all returned values from the schematizer,
        as well as a partition - message_count map (range_map)
        """
        self.log.info("Loading active topics...")
        return self._topic_to_range_map_to_topics_list(self._topics_with_messages_to_range_map)

    @cached_property
    def active_sources(self):
        """A cached dictionary of active sources, meaning that they are
        a source with an active topic.

        Output is dictionary of source_ids to
        dict of namespace_name and active_topic_count.
        """
        self.log.info("Loading active sources...")
        active_sources = {}
        for topic in self.active_topics:
            source_id = topic['source'].source_id
            if source_id in active_sources:
                active_sources[source_id]['active_topic_count'] += 1
            else:
                active_sources[source_id] = {
                    'namespace_name': topic['source'].namespace.name,
                    'active_topic_count': 1
                }
        return active_sources

    @cached_property
    def active_namespaces(self):
        """A cached dictionary of active namespaces, which are namespaces
        with at least one active source.

        Output is a dictionnary of namespace_name to dict of active_source_count
        and active_topic_count in the namespace.
        """
        self.log.info("Loading active namespaces...")
        active_namespaces = {}
        for source_id, active_source in self.active_sources.iteritems():
            namespace_name = active_source['namespace_name']
            active_topic_count = active_source['active_topic_count']
            if namespace_name in active_namespaces:
                active_namespaces[namespace_name]['active_source_count'] += 1
                active_namespaces[namespace_name]['active_topic_count'] += active_topic_count
            else:
                active_namespaces[namespace_name] = {
                    'active_source_count': 1,
                    'active_topic_count': active_topic_count
                }
        return active_namespaces

    def _create_serializable_ordered_dict_from_object_and_fields(self, obj, fields):
        result_dict = OrderedDict()
        for field in fields:
            value = getattr(obj, field)
            if isinstance(value, datetime.datetime):
                # datetimes are not json serializable
                value = str(value)
            result_dict[field] = value
        return result_dict

    def _get_topic_message_count(self, topic):
        if topic.name in self._topics_with_messages_to_range_map:
            message_count = 0
            for _, count in self._topics_with_messages_to_range_map[topic.name].iteritems():
                message_count += count
            return message_count
        return 0

    def list_topics(
        self,
        source_id=None,
        source_name=None,
        namespace_name=None,
        sort_by=None,
        descending_order=False
    ):
        if source_id:
            topics = self.schematizer.get_topics_by_source_id(
                source_id
            )
        else:
            topics = self.schematizer.get_topics_by_criteria(
                namespace_name=namespace_name,
                source_name=source_name
            )
        topics = [
            IntrospectorTopic(
                topic,
                kafka_topics=self._kafka_topics,
                topics_to_range_map=self._topics_with_messages_to_range_map
            ).to_ordered_dict() for topic in topics
        ]
        topics.sort(key=lambda topic: topic['updated_at'], reverse=True)
        if sort_by:
            topics.sort(key=lambda topic: topic[sort_by], reverse=descending_order)
        return topics

    def list_sources(
        self,
        namespace_name=None,
        sort_by=None,
        descending_order=False,
        active_sources=False
    ):
        if namespace_name:
            sources = self.schematizer.get_sources_by_namespace(namespace_name)
        else:
            self.log.info("Getting all namespaces...")
            namespaces = self.schematizer.get_namespaces()
            sources = []
            self.log.info("Getting all sources for each namespace...")
            for namespace in namespaces:
                sources += self.schematizer.get_sources_by_namespace(namespace.name)
        sources = [
            IntrospectorSource(
                source,
                active_sources=(self.active_sources if active_sources else None)
            ).to_ordered_dict() for source in sources
        ]
        sources.sort(key=lambda source: source['source_id'], reverse=True)
        if sort_by:
            sources.sort(key=lambda source: source[sort_by], reverse=descending_order)
        return sources

    def list_namespaces(
        self,
        sort_by=None,
        descending_order=False,
        active_namespaces=False
    ):
        namespaces = self.schematizer.get_namespaces()
        namespaces = [
            IntrospectorNamespace(
                namespace,
                active_namespaces=(self.active_namespaces if active_namespaces else None)
            ).to_ordered_dict() for namespace in namespaces
        ]
        namespaces.sort(key=lambda namespace: namespace['namespace_id'], reverse=True)
        if sort_by:
            namespaces.sort(key=lambda namespace: namespace[sort_by], reverse=descending_order)
        return namespaces

    def _setup_logging(self):
        CONSOLE_FORMAT = '%(asctime)s - %(name)-12s: %(levelname)-8s %(message)s'

        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(CONSOLE_FORMAT))
        handler.setLevel(logging.DEBUG)
        self.log.addHandler(handler)

    def process_args(self, args, parser):
        self.log.setLevel(max(logging.WARNING - (args.verbosity * 10), logging.DEBUG))



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from setuptools import find_packages
from setuptools import setup

import data_pipeline

readme = open('README.md').read()
doclink = """
Documentation
-------------

The full documentation is at TODO (DATAPIPE-2031|abrar): Link to public servicedocs."""
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

setup(
    # py2 + setuptools asserts isinstance(name, str) so this needs str()
    name=str('data_pipeline'),
    version=data_pipeline.__version__,
    description="Provides an interface to consume and publish to data pipeline topics.",
    long_description=readme + '\n\n' + doclink + '\n\n' + history,
    author=data_pipeline.__author__,
    author_email=data_pipeline.__email__,
    url='https://github.com/Yelp/data_pipeline',
    packages=find_packages(exclude=['tests*', 'benchmarks*']),
    include_package_data=True,
    install_requires=[
        'bravado>=8.4.0',
        'cached-property>=0.1.5',
        'cffi>=1.1.2',
        'cryptography<=1.3.4',
        'data-pipeline-avro-util>=0.2.1',
        'enum34>=1.0.4',
        'frozendict==0.5',
        'kafka-python>0.9.4,<1.0.0',
        'kafka-utils>0.3.3',
        'psutil==4.2.0',
        'pycrypto>=2.6.1',
        'pyopenssl==16.0.0',
        'pyramid_zipkin>=0.16.0',
        'PyStaticConfiguration>=0.9.0',
        'simplejson>=2.1.2',
        'swagger_zipkin>=0.1.0',
        'yelp-avro==1.9.2',
        'yelp-clog==2.2.11'
    ],
    extras_require={
        'tools': [
            # yelp_clog isn't compatible with triftpy 0.2.0, which is an
            # implicit dependency of yelp_batch
            'thriftpy<0.2.0',
            'yelp_batch>=0.19.4',
            'yelp_conn>=7.0.0',
        ],
        'testing_helpers': [
            'docker-compose==1.5.2',
            'docker-py==1.6.0',
            # requests is locked at <2.7 to satisfy a docker-compose requirement
            'requests<2.7'
        ],
        # inform downstream projects that use data_pipeline consumer to
        # include data_pipeline[internal] to their dependency.
        'internal': [
            'pysensu-yelp>=0.2.3',
            'yelp-kafka>=5.0.0',
            'yelp-servlib>=4.3.0'
        ]
    },
    zip_safe=False,
    keywords='data_pipeline',
    package_data={
        str('data_pipeline'): [
            'schemas/*.avsc',
            'testing_helpers/docker-compose.yml',
            'testing_helpers/docker-compose-opensource.yml'
        ],
    },
    scripts=[
        'bin/data_pipeline_tailer',
        'bin/data_pipeline_refresh_runner',
        'bin/data_pipeline_refresh_manager',
        'bin/data_pipeline_refresh_job',
        'bin/data_pipeline_compaction_setter',
        'bin/data_pipeline_introspector'
    ],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
)
